# Usage
# ~~~~~
# from aiko_services.main.utilities import *
#
# with ContextManager({}) as context:
#     print(context.aiko, context.message)
#
# context = ContextManager({})
# print(context.aiko)
# context.aiko["a"] = 1
#
# context = get_context()
# print(context.aiko)
#
# To Do
# ~~~~~
# - Thread local context example

# This import allows us to specify the class itself as return type,
# e.g def activate(self) -> ContextManager

from typing import Any

__all__ = ["ContextManager", "get_context"]

_CONTEXT = None


class ContextManager:
    def __init__(self, aiko: Any = None, message: Any = None) -> None:
        self.aiko = aiko
        self.message = message
        self.activate()

    def activate(self) -> "ContextManager":
        # pylint: disable=global-statement
        global _CONTEXT
        _CONTEXT = self
        return self

    def __enter__(self) -> "ContextManager":
        return self.activate()

    def __exit__(self, *args: Any) -> None:
        pass


def get_context() -> Any:
    # pylint: disable=global-statement
#   global _CONTEXT  # TODO: flake8: F824 `global _CONTEXT` is unused
    return _CONTEXT



import logging
from pathlib import Path
from typing import Dict, List, Optional, Tuple
from uuid import UUID

from highlighter.agent.capabilities._crop_processor import CropProcessor
from highlighter.agent.capabilities.base_capability import Capability, StreamEvent
from highlighter.client.gql_client import HLClient
from highlighter.client.training_runs import (
    TrainingRunArtefactType,
)
from highlighter.core.data_models import DataSample
from highlighter.predictors.onnx_yolov8 import OnnxYoloV8 as Predictor
from highlighter.predictors.output_category_descriptor import OutputCategoryDescriptor

__all__ = ["OnnxYoloV8"]

logger = logging.getLogger(__name__)


def flatten(lst):
    for x in lst:
        if isinstance(x, list):
            yield from flatten(x)
        else:
            yield x


class OnnxYoloV8(Capability):

    class InitParameters(Capability.InitParameters):
        onnx_file: Optional[str] = None
        training_run_artefact_id: Optional[UUID] = None
        num_classes: int = 80
        class_lookup: Optional[Dict[int | str, OutputCategoryDescriptor]] = None
        conf_thresh: float = 0.1
        nms_iou_thresh: float = 0.5
        nms_class_agnostic: bool = False
        is_absolute: bool = True
        crop_processor: Optional[CropProcessor] = None

    def __init__(self, context):
        super().__init__(context)
        self._predictor: Optional[Predictor] = None

        onnx_file = self.init_parameters.onnx_file
        if onnx_file is None:
            training_run_artefact_id = self.init_parameters.training_run_artefact_id
            if training_run_artefact_id is None:
                raise ValueError("Must provide `onnx_file` or `training_run_artefact_id`")
            training_run_artefact = HLClient.get_client().trainingRunArtefact(
                return_type=TrainingRunArtefactType, id=training_run_artefact_id
            )
            onnx_file = training_run_artefact.file_url

        artefact_cache_dir = Path.home() / ".cache" / "highlighter" / "agents" / "artefacts"
        # FIXME: kwargs {"device_id", "artefact_cache_dir", "onnx_file_download_timeout"}
        # should be optionally configured by the runtime context.

        self._predictor = Predictor(
            onnx_file,
            num_classes=self.init_parameters.num_classes,
            class_lookup=self.init_parameters.class_lookup,
            conf_thresh=self.init_parameters.conf_thresh,
            nms_iou_thresh=self.init_parameters.nms_iou_thresh,
            is_absolute=self.init_parameters.is_absolute,
            nms_class_agnostic=self.init_parameters.nms_class_agnostic,
            artefact_cache_dir=artefact_cache_dir,
        )

        self._crop_processor = None
        if self.init_parameters.crop_processor is not None:
            self._crop_processor = self.init_parameters.crop_processor

    def process_frame(self, stream, data_samples: List[DataSample]) -> Tuple[StreamEvent, Dict]:

        annotations = []
        if self._crop_processor is None:
            image_batch = [d.content for d in data_samples]
            batch_preds = self._predictor.predict(image_batch)

            assert len(batch_preds) == len(data_samples)
            for preds, data_sample in zip(batch_preds, data_samples):
                for pred in preds:
                    new_ann = pred.to_annotation(
                        data_sample,
                        pipeline_element_name=self.name,
                    )
                    annotations.append(new_ann)
        else:

            crops, crop_src_zones = self._crop_processor.crop_data_samples(data_samples)
            preds = self._predictor.predict(crops)

            # Translate each prediction according to the corresponding source zone
            for predictions_from_zones, zone_spec, data_sample in zip(preds, crop_src_zones, data_samples):
                for prediction in predictions_from_zones:
                    prediction.translate(*zone_spec.location.bounds[:2])

                    annotations.append(
                        prediction.to_annotation(
                            data_sample,
                            pipeline_element_name=self.name,
                        )
                    )

        entities = [a.get_or_create_entity() for a in annotations]

        return StreamEvent.OKAY, {"entities": {e.id: e for e in entities}}



import re

from pydantic import BaseModel, ConfigDict


class GQLBaseModel(BaseModel):
    """To allow for seamless integration with the GraphQL CamelCaseWorld and
    th_python_snake_case_world. We have this customized Pydantic BaseModel.

    This expects all fields are defined in snake_case.
    When serialized the resulting dict will have CamelCase keys,
    this allows it to play nice with GraphQL.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    def __init__(self, *args, **kwargs):
        kwargs = map_keys_deeply(kwargs, camel_to_snake)
        super().__init__(*args, **kwargs)

    def gql_dict(self):
        d = self.model_dump(exclude_none=True)
        d = map_keys_deeply(d, snake_to_camel)
        return d


def map_keys_deeply(x, f):
    if isinstance(x, dict):
        result = {f(k): map_keys_deeply(v, f) for k, v in x.items()}
    elif isinstance(x, list):
        result = [map_keys_deeply(element, f) for element in x]
    else:
        result = x
    return result


def camel_to_snake(s):
    if s.isupper():
        return s
    snake = re.sub("([A-Z])", r"_\g<1>", s).lower()
    if snake.startswith("_"):
        snake = snake[1:]
    return snake


def snake_to_camel(s):
    if s.isupper():
        return s
    first, *rest = s.split("_")
    return first + "".join(x.capitalize() for x in rest)



from typing import List
from urllib.parse import urlparse


def is_url_scheme(s: str, schemes: List[str]) -> bool:
    try:
        u = urlparse(s)
        return bool(u.netloc) and (u.scheme in schemes)
    except Exception as e:
        return False



import click

from ..client import ObjectClass, read_object_classes


@click.group("object-class")
@click.pass_context
def object_class_group(ctx):
    pass


def _read_object_classes(client, name: str):
    gen = read_object_classes(client, name=name)
    for obj_cls in gen:
        yield obj_cls.dict()


@object_class_group.command("read")
@click.option(
    "-i",
    "--ids",
    type=str,
    required=False,
    multiple=True,
)
@click.option(
    "-n",
    "--names",
    type=str,
    required=False,
    multiple=True,
)
@click.pass_context
def read(ctx, names, ids):
    client = ctx.obj["client"]

    result = []
    if len(ids) > 0:
        for id in ids:
            result.append(
                client.objectClass(
                    return_type=ObjectClass,
                    id=id,
                ).dict()
            )
    elif len(names) > 0:
        for name in names:
            result.extend([o for o in _read_object_classes(client, name)])

    print(result)



"""Central registry mapping *content types* to *PayloadWriter* classes.

The goal: given a **ContentTypeEnum** (and optionally a *sub‑type* string)
return *an *instance* of the corresponding writer* ready to serialise a batch
of :class:`DataSample`s.

A writer must expose two public attributes/methods:

* ``write(samples) -> bytes`` – serialise iterable of ``DataSample`` → bytes

See ``highlighter/io/writers.py`` for concrete implementations such as
:class:`VideoWriter`, :class:`ImageWriter`, or :class:`TextWriter`.
"""

# TODO:
# import logging
from typing import Callable, Dict, Tuple, Type

from highlighter.core.enums import ContentTypeEnum
from highlighter.io.base import PayloadWriter  # the Protocol from earlier

# TODO:
# logger = logging.getLogger(__name__)

# Key is ContentTypeEnum, however we should consider adding subtype (which  may be None) eg., Tuple[ContentTypeEnum, str | None]
_RegistryKey = Tuple[ContentTypeEnum]

# Internal mapping → writer *class* (not instance)
_writer_by_type: Dict[_RegistryKey, Type[PayloadWriter]] = {}


# --------------------------------------------------------------------------- #
# Public API
# --------------------------------------------------------------------------- #
def register_writer(
    content_type: ContentTypeEnum,
    writer_cls: Type[PayloadWriter],
    *,
    force: bool = False,
    # subtype: str | None = None, # TODO: consider adding
) -> None:
    """Register *writer_cls* for *content_type*.

    Parameters
    ----------
    content_type
        Enum value like ``ContentTypeEnum.IMAGE``.
    writer_cls
        A class implementing the :class:`PayloadWriter` protocol.
    force
        If False (default) and a writer is already registered for the key, a
        ``ValueError`` is raised.  If True, silently overwrites.
    """
    key: _RegistryKey = content_type
    if not force and key in _writer_by_type:
        raise ValueError(f"Writer already registered for {key}")
    _writer_by_type[key] = writer_cls
    # TODO:
    # logger.debug("Registered writer %s for %s", writer_cls.__name__, key)


def get_writer(
    content_type: ContentTypeEnum,
    **kwargs,
) -> PayloadWriter:
    """Instantiate and return the writer for *content_type*.

    ``kwargs`` are forwarded to the writer's constructor.
    """
    key: _RegistryKey = content_type
    try:
        writer_cls = _writer_by_type[key]
    except KeyError as exc:
        available = ", ".join(str(k) for k in _writer_by_type)
        raise ValueError(f"No writer registered for key {key}. Available: {available}") from exc
    return writer_cls(**kwargs)


# --------------------------------------------------------------------------- #
# Default built‑in registrations
# --------------------------------------------------------------------------- #
# Import *after* function definitions to avoid circular dependencies when the
# writers themselves import this registry for registration-at-import‑time.

try:
    from highlighter.io.writers import (  # CSVWriter,
        EntityWriter,
        ImageWriter,
        TextWriter,
        VideoWriter,
    )
except ModuleNotFoundError:
    raise
    # The host project may choose to provide its own writers – log & continue.
    # TODO:
    # logger.warning("Default writer implementations not found; registry is empty")
else:
    # Basic image/video/text mappings
    register_writer(ContentTypeEnum.IMAGE, ImageWriter)
    register_writer(ContentTypeEnum.VIDEO, VideoWriter)
    register_writer(ContentTypeEnum.TEXT, TextWriter)
    register_writer(ContentTypeEnum.ENTITIES, EntityWriter)
    # register_writer(ContentTypeEnum.CSV, CSVWriter)

    # TODO: Consider supporting specialised subtypes
    # eg., register_writer(ContentTypeEnum.TEXT, CSVWriter, subtype="csv")



from .const import *
from .decorators import *
from .geometry import *
from .gql_base_model import *
from .itertools import *
from .labeled_uuid import *
from .memory_cleanup import *
from .pagination import *
from .utilities import *
from .visualize import *



# Usage
# -----
# from aiko_services.main.utilities import *
# module_descriptor = "pathname/filename.py"  # or "package.module"
# module = load_module(module_descriptor)
# module.some_class()
# module.some_function()
#
# To Do
# -----
# - None, yet !

import importlib
import os
import sys

__all__ = ["load_module", "load_modules"]

if os.environ.get("AIKO_IMPORTER_USE_CURRENT_DIRECTORY"):
    sys.path.append(os.getcwd())

MODULES_LOADED = {}

def load_module(module_descriptor):
    if module_descriptor in MODULES_LOADED:
        module = MODULES_LOADED[module_descriptor]
    else:
        if module_descriptor.endswith(".py"):
            # Load module from Python source pathname, e.g "directory/file.py"
            module_pathname = module_descriptor
            module = importlib.machinery.SourceFileLoader(
                'module', module_pathname).load_module()
        else:
            # Load module from "installed" modules, e.g "package.module"
            module_name = module_descriptor
            module = importlib.import_module(module_name)
        MODULES_LOADED[module_descriptor] = module
    return module

def load_modules(module_pathnames):
    modules = []
    for module_pathname in module_pathnames:
        if module_pathname:
            modules.append(load_module(module_pathname))
        else:
            modules.append(None)
    return modules



# To Do
# ~~~~~
# - Provide Aiko Dashboard metrics and support Metrics
#
# - Support media types "text" and "text/zip" ?
#
# - Support ZMQ REQ/REP for bidirectional request / response for control flow
#   - Provide confirmation back to the ZMQ client (zmq.REQ/REP)

import queue
from threading import Thread
import zmq

import aiko_services as aiko
from aiko_services.main.utilities import get_network_port_free

__all__ = ["DataSchemeZMQ"]

_LOGGER = aiko.process.logger(__name__)

# --------------------------------------------------------------------------- #
# parameter: "data_sources" provides the ZMQ server bind details (incoming)
# - "data_sources" list should only contain a single entry
# - "(zmq://hostname:port_range)"  port_range may be a single port or a range
#   - "(zmq://*:*)"                any available TCP port      (only localhost)
#   - "(zmq://*:6502)"             a given TCP port            (only localhost)
#   - "(zmq://*:6502-6510)"        any TCP port in the specified range
#   - "(zmq://localhost:6502)"     given hostname and TCP port
#   - "(zmq://0.0.0.0:*)"          any available TCP port      (from any host)
#   - "(zmq://0.0.0.0:6502)"       a given TCP port            (from any host)
#   - "(zmq://0.0.0.0:6502-6510)"  any TCP port in the specified range
#
# parameter: "data_targets" provides the ZMQ client connect details (outgoing)
# - "data_targets" list should only contain a single entry
# - "(zmq://hostname:port)" ...
#   - "(zmq://*:6502)"             localhost and TCP port
#   - "(zmq://localhost:6502)"     localhost and TCP port
#   - "(zmq://0.0.0.0:6502)"       given hostname and TCP port

class DataSchemeZMQ(aiko.DataScheme):
    def create_sources(self,
        stream, data_sources, frame_generator=None, use_create_frame=False):

        if not frame_generator:
            frame_generator = self.frame_generator

        try:
            zmq_url = self._parse_zmq_url(data_sources, find_free_port=True)
        except ValueError as value_error:
            return aiko.StreamEvent.ERROR, {"diagnostic": value_error}
        self.share["zmq_url"] = zmq_url
        _LOGGER.debug(f"create_sources(): zmq_url: {zmq_url}")

        self.zmq_context = zmq.Context()
        self.zmq_socket = self.zmq_context.socket(zmq.PULL)
        self.zmq_socket.bind(zmq_url)  # TODO: Exception handling
        self.zmq_socket.setsockopt(zmq.RCVTIMEO, 1000)

        self.queue = queue.Queue()
        self.terminate = False
        Thread(target=self._run, daemon=True).start()
        self.pipeline_element.create_frames(stream, frame_generator)
        return aiko.StreamEvent.OKAY, {}

    def create_targets(self, stream, data_targets):
        try:
            zmq_url = self._parse_zmq_url(data_targets)
        except ValueError as value_error:
            return aiko.StreamEvent.ERROR, {"diagnostic": value_error}
        self.share["zmq_url"] = zmq_url
        _LOGGER.debug(f"create_targets(): zmq_url: {zmq_url}")

        self.zmq_context = zmq.Context()
        self.zmq_socket = self.zmq_context.socket(zmq.PUSH)
        self.zmq_socket.connect(zmq_url)  # TODO: Exception handling

        stream.variables["target_zmq_socket"] = self.zmq_socket
        return aiko.StreamEvent.OKAY, {}

    def destroy_sources(self, stream):
        self.terminate = True

    def destroy_targets(self, stream):
        self._zmq_destroy()

    def frame_generator(self, stream, frame_id):
        data_batch_size, _ = self.pipeline_element.get_parameter(
            "data_batch_size", default=1)
        data_batch_size = int(data_batch_size)

        records = []
        while (data_batch_size > 0):
            data_batch_size -= 1
            if not self.queue.qsize():
                break
            record = self.queue.get()
            records.append(record)

        if records:
            return aiko.StreamEvent.OKAY, {"records": records}
        else:
            return aiko.StreamEvent.NO_FRAME, {}

    def _parse_zmq_url(self, data_urls, find_free_port=False):
        data_url = data_urls[0]
        path = aiko.DataScheme.parse_data_url_path(data_url)
        tokens = path.split(":")  # "hostname:port_range", e.g "*:6502-6510"
        diagnostic =  \
            f'ZMQ Data URL "{data_url}" must be "zmq://host:port_range"'
        if len(tokens) != 2:
            raise ValueError(diagnostic)
        hostname = tokens[0]

        port_range = tokens[1].split("-")
        if len(port_range) == 1:
            port_range[0] = "0" if port_range[0] == "*" else port_range[0]
            port_range = [port_range[0], port_range[0]]
        for index, port_number in enumerate(port_range):
            if not port_number.isdigit():
                raise ValueError(diagnostic)
            port_range[index] = int(port_number)
        if find_free_port and port_range[0] != port_range[1]:
            port = get_network_port_free(port_range)
        else:
            port = port_range[0]

        return f"tcp://{hostname}:{port}"

    def _run(self):
        try:
            while not self.terminate:
                try:
                    payload = self.zmq_socket.recv()
                    self.queue.put(payload)
                except zmq.Again:
                    continue
        finally:
            self._zmq_destroy()

    def _zmq_destroy(self):
        if self.zmq_socket:
            self.zmq_socket.close()
            self.zmq_socket = None
        if self.zmq_context:
            self.zmq_context.term()
            self.zmq_context = None

aiko.DataScheme.add_data_scheme("zmq", DataSchemeZMQ)

# --------------------------------------------------------------------------- #



from __future__ import annotations

import logging
from collections import deque
from dataclasses import dataclass, field
from datetime import UTC, datetime, timedelta
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
from uuid import UUID, uuid4

from highlighter.client.assessments import finalise
from highlighter.client.base_models.base_models import CaseType, TaskType
from highlighter.client.gql_client import HLClient, get_threadsafe_hlclient
from highlighter.core.gql_base_model import GQLBaseModel

if TYPE_CHECKING:
    from highlighter.agent.capabilities.recorder import Recorder

logger = logging.getLogger(__name__)

__all__ = [
    "update_task_status",
    "update_task",
    "lease_task",
    "lease_tasks_from_steps",
    "add_files_to_order",
]


class TaskStatus(str, Enum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    SUCCESS = "SUCCESS"

    @staticmethod
    def validate_str(s) -> bool:
        return s in [s.value for s in TaskStatus]


class UpdateTaskResultPayload(GQLBaseModel):
    submission: CreateSubmissionNotFinalisedPayload.SubmissionType
    errors: List[Any]


class DataFile(GQLBaseModel):
    id: Optional[str] = None
    uuid: Optional[str] = None
    original_source_url: Optional[str] = None
    file_url_original: Optional[str] = None
    content_type: Optional[str] = None


class Case(GQLBaseModel):
    class CaseSubmission(GQLBaseModel):
        id: str
        uuid: UUID
        data_files: List[DataFile]

    id: str
    latest_submission: Optional[CaseSubmission] = None
    entity_id: Optional[str] = None
    data_files: Optional[List[DataFile]] = None


class Task(GQLBaseModel):
    id: str
    status: Optional[TaskStatus] = None
    case: Optional[Case] = None
    leased_until: Optional[datetime] = None
    parameters: Optional[Dict[str, Any]] = None


def update_task_status(
    client: HLClient,
    task_id: str,
    status: Union[str, TaskStatus],
    message: Optional[str] = None,
):

    assert isinstance(status, TaskStatus) or TaskStatus.validate_str(status), f"Got: {status}"

    class UpdateTaskStatusResponse(GQLBaseModel):
        errors: List[Any]

    kwargs = {
        "id": task_id,
        "status": status,
    }
    if message is not None:
        kwargs["message"] = message

    response = client.update_task_status(return_type=UpdateTaskStatusResponse, **kwargs)
    if response.errors:
        raise ValueError(f"{response.errors}")

    return response


def update_task(
    client: HLClient,
    task_id: Union[UUID, str],
    status: Optional[Union[str, TaskStatus]] = None,
    leased_until: Optional[Union[datetime, str]] = None,
    lease_sec: Optional[int] = None,
    **kwargs,
) -> Task:

    if lease_sec:
        assert leased_until is None, "Cannot use both leased_until and lease_sec"
        leased_until = (datetime.now(UTC) + timedelta(seconds=lease_sec)).isoformat()

    if isinstance(leased_until, datetime):
        leased_until = leased_until.isoformat()

    class TaskResponse(GQLBaseModel):
        task: Task
        errors: Any

    kwargs = dict(
        id=str(task_id),
        status=status,
        leasedUntil=leased_until,
        **kwargs,
    )

    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    response = client.updateTask(return_type=TaskResponse, **kwargs)

    if response.errors:
        raise ValueError(f"Errors: {response.errors}")

    return response.task


def lease_task(
    client: HLClient,
    task_id: Union[UUID, str],
    set_status_to: Optional[Union[str, TaskStatus]] = None,
    leased_until: Optional[Union[datetime, str]] = None,
    lease_sec: Optional[float] = None,
) -> Task:

    if lease_sec:
        assert leased_until is None, "Cannot use both leased_until and lease_sec"
        leased_until = (datetime.now(UTC) + timedelta(seconds=lease_sec)).isoformat()

    if isinstance(leased_until, datetime):
        leased_until = leased_until.isoformat()

    class TaskResponse(GQLBaseModel):
        task: Task
        errors: Any

    update_task_args = {"id": str(task_id), "leasedUntil": leased_until}
    if set_status_to is not None:
        update_task_args["status"] = set_status_to
    response = client.updateTask(return_type=TaskResponse, **update_task_args)
    if response.errors:
        raise ValueError(f"Errors: {response.errors}")
    return response.task


def lease_tasks_from_steps(
    client: HLClient,
    step_ids: List[Union[UUID, str]],
    count: int = 1,
    filter_by_status: Optional[Union[str, TaskStatus]] = None,
    set_status_to: Optional[Union[str, TaskStatus]] = None,
    leased_until: Optional[Union[datetime, str]] = None,
    lease_sec: Optional[float] = None,
    filter_by_task_id: Optional[List[UUID]] = None,
) -> List[Task]:

    if filter_by_status:
        assert isinstance(filter_by_status, TaskStatus) or TaskStatus.validate_str(filter_by_status)

    if set_status_to:
        assert isinstance(set_status_to, TaskStatus) or TaskStatus.validate_str(set_status_to)

    if lease_sec:
        assert leased_until is None, "Cannot use both leased_until and lease_sec"
        leased_until = (datetime.now(UTC) + timedelta(seconds=lease_sec)).isoformat()

    if isinstance(leased_until, datetime):
        leased_until = leased_until.isoformat()

    class TaskResponse(GQLBaseModel):
        errors: List[Any]
        tasks: List[Task]

    response = client.leaseTasksFromSteps(
        return_type=TaskResponse,
        stepIds=[str(s) for s in step_ids],
        count=count,
        leasedUntil=leased_until,
    )

    if len(response.errors) > 0:
        raise ValueError(response.errors)

    if filter_by_status:
        tasks = [t for t in response.tasks if t.status == filter_by_status]
    else:
        tasks = response.tasks

    if filter_by_task_id:
        task_ids = []

        for i in filter_by_task_id:
            if isinstance(i, str):
                task_ids.append(i)
            elif isinstance(i, UUID):
                task_ids.append(str(i))
            else:
                raise ValueError()

        tasks = [t for t in tasks if t.id in task_ids]

    if set_status_to:
        for task in tasks:
            _ = update_task_status(client, task.id, set_status_to)
    return tasks


def add_files_to_order(client: HLClient, order_id: str, file_ids: List[str]) -> Dict[str, str]:
    """
    add files to a workflow order, creating tasks for them.
    returns a dictionary mapping data_file_id to task_id.
    """

    class AddFilesResponse(GQLBaseModel):
        tasks: List[Task]
        errors: List[Any]

    # this mutation is dynamically dispatched by HLClient
    res = client.addFilesToWorkflowOrder(
        return_type=AddFilesResponse, workflowOrderId=order_id, fileIds=file_ids
    )

    if res.errors:
        raise ValueError(f"Failed to add files to order: {res.errors}.")

    file_to_task = {}
    if res.tasks:
        for t in res.tasks:
            if t.case and t.case.data_files:
                for df in t.case.data_files:
                    # map the file id (not uuid) to the task id
                    file_to_task[str(df.id)] = t.id

    return file_to_task


# FIXME: (Josh). Should the be merged with TaskType? We have 2 modes of using Task
# one for Gql and the other for Agents. I think they should be one-in-the-same.


@dataclass
class ProcessorState:
    """Lock-free coordination record shared between TaskContext and Recorder workers."""

    processor: "Recorder"  # forward-declared to avoid import cycle
    submission: CreateSubmissionNotFinalisedPayload.SubmissionType
    case_id: Optional[str] = None  # Case ID for logging and tracking
    pending_chunks: int = 0
    stopping: bool = False
    finished: bool = False
    error: Optional[BaseException] = None


@dataclass
class RecordingSession:
    """Tracks the processors contributing to a submission and their progress."""

    submission: CreateSubmissionNotFinalisedPayload.SubmissionType
    case_id: str
    processors: List[ProcessorState] = field(default_factory=list)

    def all_finished(self) -> bool:
        return all(p.finished for p in self.processors)

    def has_error(self) -> bool:
        return any(p.error is not None for p in self.processors)

    def iter_errors(self):
        for state in self.processors:
            if state.error is not None:
                yield state


class TaskContext:

    def __init__(
        self,
    ):
        self._recording = deque()
        self.logger = logging.getLogger(__name__)
        self._hl_credentials: Optional[Tuple[str, str]] = None

    def _get_threadsafe_hl_client(self) -> HLClient:
        """Lazily capture credentials and return a thread-local HLClient."""
        if self._hl_credentials is None:
            base_client = HLClient.get_client()
            self._hl_credentials = (base_client.api_token, base_client.endpoint_url)

        api_token, endpoint_url = self._hl_credentials
        return get_threadsafe_hlclient(api_token, endpoint_url)

    def start_recording(
        self,
        dsps_to_record,
        workflow_order_id: UUID,
        case_name: Optional[str] = None,
        entity_id: Optional[UUID] = None,
        *,
        log_context: Optional[str] = None,
    ):
        client = self._get_threadsafe_hl_client()
        case, sub = create_empty_case_with_not_finalized_submission(
            client,
            workflow_order_id,
            case_name=case_name,
            entity_id=entity_id,
            log_context=log_context,
        )
        session = RecordingSession(submission=sub, case_id=case.id)
        for dsp in dsps_to_record:
            processor_state = ProcessorState(processor=dsp, submission=sub, case_id=case.id)
            session.processors.append(processor_state)
            dsp.start_recording(sub, processor_state)

        prefix = f"{log_context} " if log_context else ""
        logger.info(
            f"{prefix}Start Recording: not_finalised_submission: {sub.id}, case: {case.id}",
            extra={"color": "green"},
        )
        self._recording.append(session)

    def finalise_submissions_on_recording_state_off(self):
        """Check for completed recordings and finalize their submissions.

        This method polls the shared ProcessorState records for the oldest recording session.
        Once every Recorder reports finished=True the submission is finalized.
        """
        if not self._recording:
            return

        recording_session = self._recording[0]

        # Surface any worker-thread errors to the runtime loop.
        for state in recording_session.iter_errors():
            self.logger.error(
                "Recording processor %s failed while handling submission %s: %s",
                state.processor,
                recording_session.submission.id,
                state.error,
            )
            # Propagate the first failure.
            raise state.error

        if not recording_session.all_finished():
            return

        submission = recording_session.submission
        self.logger.info(f"Finalized not_finalised_submission: {submission.id}")
        client = self._get_threadsafe_hl_client()
        finalise(client, submission)

        # Let each processor clear any per-submission state.
        for state in list(recording_session.processors):
            try:
                state.processor.clear_completed_recording(state)
            except Exception as exc:  # pragma: no cover - defensive
                self.logger.warning(
                    "Failed to clear recording state for processor %s (%s): %s",
                    state.processor,
                    submission.id,
                    exc,
                )

        self._recording.popleft()

    def stop_recording(self):

        if not self._recording:
            return

        recording_session = self._recording[0]
        for state in list(recording_session.processors):
            dsp = state.processor
            self.logger.info(
                f"Stop Recording: not_finalised_submission: {recording_session.submission.id}, case: {recording_session.case_id}",
                extra={
                    "stream_id": getattr(dsp, "_stream_id", None),
                    "capability_name": getattr(dsp, "_capability_name", None),
                },
            )
            dsp.stop_recording(state)


from highlighter.core.gql_base_model import GQLBaseModel


class CreateCasePayload(GQLBaseModel):
    errors: List[str]
    case: Optional[CaseType] = None


class CreateSubmissionNotFinalisedPayload(GQLBaseModel):

    class SubmissionType(GQLBaseModel):
        id: str
        case_id: Optional[str] = None
        task_id: Optional[str] = None

    submission: Optional[SubmissionType] = None
    errors: List[str]


class CreateTaskPayload(GQLBaseModel):
    errors: List[str]
    task: Optional[TaskType] = None


def create_empty_case_with_not_finalized_submission(
    client: HLClient,
    workflow_order_id: UUID,
    case_name: Optional[str] = None,
    entity_id: Optional[UUID] = None,
    *,
    log_context: Optional[str] = None,
) -> Tuple[CaseType, CreateSubmissionNotFinalisedPayload.SubmissionType]:
    """Creates an empty Case

    Args:
        client: HLClient instance
        workflow_order_id: Target workflow order ID
        case_name: Optional name for the case
        entity_id: Optional entity ID

    Returns:
        Case object

    Raises:
        RuntimeError: If case creation fails
    """
    # Create the case
    case_result = client.createCase(
        return_type=CreateCasePayload,
        workflowOrderId=workflow_order_id,
        name=case_name,
        entityId=entity_id,
        state="ready",
    )

    if case_result.errors:
        raise RuntimeError(f"Error in createCase: {case_result.errors}")

    if not case_result.case:
        raise RuntimeError("createCase succeeded but no CaseType returned")

    prefix = f"{log_context} " if log_context else ""
    logger.info(f"{prefix}Created case {case_result.case.id}", extra={"color": "green"})
    case = case_result.case

    sub_result = client.createSubmissionNotFinalised(
        return_type=CreateSubmissionNotFinalisedPayload,
        caseId=case.id,
    )

    if sub_result.errors:
        raise RuntimeError(f"Error in createSubmissionNotFinalised: {sub_result.errors}")

    if not sub_result.submission:
        raise RuntimeError(
            "createSubmissionNotFinalised succeeded but no CreateSubmissionNotFinalisedPayload.SubmissionType returned"
        )

    sub = sub_result.submission

    return case, sub



# Usage
#
# AIKO_NAMESPACE=test
# AIKO_MQTT_HOST=localhost
# AIKO_MQTT_PORT=1883      # Or 1884 (websockets), 9883 (TLS), 9884 (WSS)
# AIKO_MQTT_TLS=true       # If unspecified then depends on AIKO_USERNAME value
# AIKO_MQTT_TRANSPORT=tcp  # Or "websockets"
# AIKO_USERNAME=username   # By default, specifying a username enables TLS (SSL)
# AIKO_PASSWORD=password
#
# mqtt_configuration = get_mqtt_configuration(tls_enabled=True)
#
# Where mqtt_configuration is a tuple ...
#     (mqtt_host, mqtt_port, mqtt_transport, username, password, tls_enabled)
#
# Resources
# ~~~~~~~~~
# https://stackoverflow.com/questions/24196932/how-can-i-get-the-ip-address-of-eth0-in-python
#
# To Do
# ~~~~~
# - Replace hard-coded "_AIKO_MQTT_HOSTS" with environment variable
# - Move discovery and bootstrap functions into "message/discovery.py"
# - Implement discovery for finding the default MQTT hostname
# - Implement discovery for finding the default namespace
#   - Define Namespace class for holding information about a specific namespace
#
# To Do: Tests
# ~~~~~~~~~~~~
# unset AIKO_MQTT_HOST  # unconfigured
#                       # --> try "localhost"
# export AIKO_MQTT_HOST=no_such_host.com  # DNS host name or service unknown
#                       # --> try and fail during MQTT.connect()
# export AIKO_MQTT_HOST=valid_host.com    # No MQTT server (broker)
#                       # --> try and fail during MQTT.connect()

import getpass
import os
import secrets
import socket
from threading import Thread
import time

from .logger import get_logger

__all__ = [
    "create_password",
    "get_hostname", "get_mqtt_configuration", "get_mqtt_host", "get_mqtt_port",
    "get_namespace", "get_namespace_prefix", "get_pid", "get_username"
]

_AIKO_BOOTSTRAP_UDP_PORT = 4149
_AIKO_MQTT_HOSTS = []  # List of host (name, port) to check for MQTT server
_AIKO_MQTT_HOST = "localhost"
_AIKO_MQTT_PORT = 1883        # TCP/IP: 9883, WebSockets: 9884
_AIKO_MQTT_TRANSPORT = "tcp"  # "websockets"
_AIKO_NAMESPACE = "aiko"

_LOCALHOST_IP = "127.0.0.1"
_LOGGER = get_logger(__name__)

def create_password(length=32):
    return secrets.token_hex(length)

def _get_lan_ip_address():
    try:
        ip_address = ((
                [ip for ip in socket.gethostbyname_ex(socket.gethostname())[2]
                     if not ip.startswith("127.")
                ]
                or
                [[(s.connect(("8.8.8.8", 53)), s.getsockname()[0], s.close())
                    for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]
                ][0][1]]
            ) + [_LOCALHOST_IP])[0]
    except Exception:  # typically this will be "socket.gaierror"
        _LOGGER.warning('Aiko Services using "localhost" as your hostname')
        ip_address = _LOCALHOST_IP
    return ip_address

def _host_server_up(host, port):
    try:
        _socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        _socket.connect((host, port))
        _socket.close()
        result = True
    except socket.error as exception:
        result = False
    return result

def get_hostname():
    hostname = socket.gethostname()
    if hostname.find(".") < 0 and hostname == "localhost":
        hostname = socket.gethostbyaddr(hostname)[0]
    if hostname.endswith("amazonaws.com"):  # shorten AWS EC2 hostname
        hyphen = hostname.find("-") + 1
        fullstop = hostname.find(".")
        hostname = hostname[hyphen:fullstop].replace("-", ".")
    return hostname.removesuffix(".local")

def get_mqtt_configuration(tls_enabled=None):
    server_up, mqtt_host, mqtt_port = get_mqtt_host()
    mqtt_transport = os.environ.get("AIKO_MQTT_TRANSPORT", _AIKO_MQTT_TRANSPORT)
    username = os.environ.get("AIKO_USERNAME", None)
    password = os.environ.get("AIKO_PASSWORD", None)

    if tls_enabled == None:
        mqtt_tls = os.environ.get("AIKO_MQTT_TLS", None)
        if mqtt_tls:
            tls_enabled = mqtt_tls == "true"
        else:
            tls_enabled = (username is not None) and (len(username) > 0)
    return (server_up, mqtt_host, mqtt_port,  \
            mqtt_transport, username, password, tls_enabled)

# Try in order ...
# - Environment variables: AIKO_MQTT_HOST and AIKO_MQTT_PORT
# - _AIKO_MQTT_HOSTS list
# - _AIKO_MQTT_HOST and _AIKO_MQTT_PORT

def get_mqtt_host():
    mqtt_hosts = _AIKO_MQTT_HOSTS.copy()

    mqtt_host = os.environ.get("AIKO_MQTT_HOST", None)
    mqtt_port = get_mqtt_port()
    if mqtt_host:
        mqtt_hosts.insert(0, (mqtt_host, mqtt_port))

    mqtt_hosts.append((_AIKO_MQTT_HOST, mqtt_port))

    server_up = False
    for host, port in mqtt_hosts:
        time_start = time.time()
        if _host_server_up(host, port):
            server_up = True
            mqtt_host = host
            mqtt_port = port
            break
        delay = time.time() - time_start
        _LOGGER.warning(f"MQTT host {host} timeout after {delay:.02f} seconds")
    return server_up, mqtt_host, mqtt_port

def get_mqtt_port():
    return int(os.environ.get("AIKO_MQTT_PORT", _AIKO_MQTT_PORT))

def get_namespace():
    return os.environ.get("AIKO_NAMESPACE", _AIKO_NAMESPACE)

def get_namespace_prefix():
    namespace_prefix = ""
    namespace = get_namespace()
    if ":" in namespace:
        namespace_prefix = namespace[0:namespace.find(":") + 1]
    return namespace_prefix

def get_pid():
    return str(os.getpid())

def get_username():
    return getpass.getuser()

# Bootstrap devices that don't support DNS or mDNS yet.
# Simple UDP broadcast request / unicast reply to find the MQTT server.
# Request  message: "boot? response_ip_address response_ip_port"
# Response message: "boot mqtt_ip_address mqtt_ip_port namespace"

RESPONSE = "boot " +  \
    _get_lan_ip_address() + " " + str(get_mqtt_port()) + " " + get_namespace()

def bootstrap_thread():
    time.sleep(1)  # wait for previous service manager to terminate
    _socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        _socket.bind(("0.0.0.0", _AIKO_BOOTSTRAP_UDP_PORT))
        while True:
            message, address = _socket.recvfrom(256)
            message = message.decode("utf-8")
            tokens = message.split()
            if len(tokens) == 3  and tokens[0] == "boot?":
                _LOGGER.info(f"Bootstrap request: {tokens[1]}:{tokens[2]}")
                _socket.sendto(RESPONSE.encode(), (tokens[1], int(tokens[2])))
    except Exception as exception:
        _LOGGER.warning(f"Bootstrap thread stopped: {exception}")

def bootstrap_start():
    thread = Thread(target=bootstrap_thread, args=())
    thread.daemon = True
    thread.start()



"""
Data payload writers for various content types.

Write Modes:
-----------
Writers support different modes depending on the output format:

Streaming (incremental) writes:
    Data is written incrementally as it arrives, minimizing memory usage.
    Supported by: VideoWriter (for video formats like MP4, AVI)
    Use when: Processing long-duration content or continuous data streams

Batch (buffered) writes:
    All data is buffered in memory before encoding.
    Required by: Formats that need the complete dataset (e.g., ImageWriter for
    multi-image grids, EntityWriter for Avro aggregation)
    Use when: Data size is manageable or format requires full dataset upfront

Note: Not all writers support streaming. Check individual writer documentation
for supported modes.
"""

import io
import logging
import math
import os
from dataclasses import dataclass
from fractions import Fraction
from pathlib import Path
from typing import BinaryIO, Iterable, Union

import av
import numpy as np
from PIL import Image

from highlighter.agent.utilities import EntityAggregator, FileAvroEntityWriter
from highlighter.client import ENTITY_AVRO_SCHEMA
from highlighter.client.base_models.entities import Entities
from highlighter.core.data_models.data_sample import DataSample
from highlighter.io.base import PayloadWriter

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class EncodeSettings:
    codec: str = "h264"
    pix_fmt: str = "yuv420p"
    crf: int = 23
    preset: str = "medium"
    scenecut: str = "0"


@dataclass
class StreamingState:
    """State for streaming video writes.

    Encapsulates all stateful information needed for incremental video encoding.
    """

    container: av.container.Container | None = None
    stream: av.stream.Stream | None = None
    sink: BinaryIO | None = None
    close_sink: bool = False
    width: int | None = None
    height: int | None = None
    is_open: bool = False

    def reset(self) -> None:
        """Reset all state to initial values."""
        self.container = None
        self.stream = None
        self.sink = None
        self.close_sink = False
        self.width = None
        self.height = None
        self.is_open = False


class ImageWriter(PayloadWriter):
    """
    Serialise one or many image samples as a single **PNG**.

    * If exactly one sample is supplied, the payload is just that image.
    * If more than one sample is supplied, the images are tiled into a
      square-ish grid:  ``cols = ceil(sqrt(N)), rows = ceil(N/cols)``.
    """

    def __init__(self, *, mode: str = "RGB", extension: str = "PNG"):
        """
        Parameters
        ----------
        mode
            PIL/Pillow image mode for the final canvas (defaults to ``"RGB"``).
        extension
            PIL/Pillow image extension for the final canvas. Required if `write`ing
            to a `sink` of type `BinaryIO`. If `sink` is a `str` or `os.PathLike`
            then `extension is ignored. (defaults to ``"PNG"``).


        """
        self.mode = mode
        self.extension = extension

    # The signature mirrors VideoWriter so DataFile.save_local can treat every
    # writer uniformly.
    def write(
        self,
        samples: Iterable[DataSample],
        sink: Union[str, os.PathLike, BinaryIO],
    ) -> None:
        samples = list(samples)
        if not samples:
            raise ValueError("ImageWriter.write() received no samples")

        # Convert every DataSample → PIL.Image
        pil_images = []
        for s in samples:
            if isinstance(s.content, Image.Image):
                img = s.content.convert(self.mode)
            else:
                arr = s.to_ndarray()
                if arr.ndim == 2:  # greyscale → RGB
                    arr = np.stack([arr] * 3, axis=-1)
                img = Image.fromarray(arr).convert(self.mode)
            pil_images.append(img)

        # Tile if necessary
        if len(pil_images) == 1:
            final_img = pil_images[0]
        else:
            w, h = pil_images[0].size
            n = len(pil_images)
            cols = math.ceil(math.sqrt(n))
            rows = math.ceil(n / cols)
            canvas = Image.new(self.mode, (cols * w, rows * h))
            for idx, img in enumerate(pil_images):
                r, c = divmod(idx, cols)
                canvas.paste(img.resize((w, h)), (c * w, r * h))
            final_img = canvas

        # Write to the requested sink
        if isinstance(sink, (str, Path)):
            with open(sink, "wb") as f:
                final_img.save(f)
        else:
            final_img.save(sink, format=self.extension)


class TextWriter(PayloadWriter):
    def write():
        pass


class EntityWriter(PayloadWriter):
    def __init__(self, *, extension: str = "avro"):
        """
        Parameters
        ----------
        extension
            file extension which defaults to avro
        """
        self.extension = extension

    def write(
        self,
        samples: Iterable[DataSample],
        sink: Union[str, os.PathLike, BinaryIO],
    ) -> None:
        """
        Encode all Entities into *sink*
        """
        writer = FileAvroEntityWriter(
            ENTITY_AVRO_SCHEMA,
            sink,
        )
        agg = EntityAggregator(
            minimum_track_frame_length=3,
            minimum_embedding_in_track_frame_length=3,
            writer=writer,
        )
        samples = list(samples)
        for sample in samples:
            if not isinstance(sample.content, Entities):
                raise ValueError("EntityWriter requires Entities sample content")
            agg.append_entities(list(sample.content))
        agg.write()


class VideoWriter(PayloadWriter):
    """
    Stream-encode frames into a video sink with efficient, C-based resizing.

    Features:
      • Configurable frame_rate, bit_rate, resolution
      • Pre-flight probing for resolution and bitrate
      • `av.VideoFrame.reformat` for on-the-fly scaling (fast C)
      • Supports both batch and streaming write modes

    Write Modes:
    -----------
    Streaming mode (incremental writes):
        Write frames one at a time to minimize memory usage. Ideal for
        long-running video captures where frames arrive continuously.

        Usage:
            with VideoWriter(frame_rate=24.0) as writer:
                writer.open(sink_path, first_sample)
                for sample in samples:
                    writer.write_frame(sample)
                # Automatically closes on exit

    Batch mode (buffer all samples):
        Buffer all frames in memory then encode. Simpler API for small videos
        or when all frames are already in memory.

        Usage:
            writer = VideoWriter(frame_rate=24.0)
            writer.write(samples, sink_path)

    When to use which mode:
    ----------------------
    - Use streaming: Long-duration videos or continuous captures to minimize memory
    - Use batch: Short videos where all frames fit comfortably in memory
    """

    supports_resize = True

    def __init__(
        self,
        *,
        frame_rate: float = 24.0,
        bit_rate: int | None = None,
        resolution: tuple[int, int] | None = None,
        extension: str = "mp4",
        settings: EncodeSettings = EncodeSettings(),
    ):
        self.frame_rate = frame_rate
        self.bit_rate = bit_rate
        self.resolution = resolution
        self.settings = settings
        self.extension = extension

        # Streaming state encapsulated in dataclass
        self._state = StreamingState()

    def __enter__(self):
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit - ensures container is closed."""
        if self._state.is_open:
            self.close()
        return False

    def _prepare(self, first_frame: DataSample) -> tuple[int, int]:
        """
        Determine target width/height and estimate bit_rate if unset.
        """
        arr = first_frame.to_ndarray()
        h, w = arr.shape[:2]
        if self.resolution:
            width, height = self.resolution
        else:
            width, height = w, h
        if self.bit_rate is None:
            raw_bytes = arr.nbytes
            # use 25% of raw size as heuristic
            self.bit_rate = int(raw_bytes * self.frame_rate * 8 * 0.25)
        return width, height

    def open(
        self,
        sink: Union[str, os.PathLike, BinaryIO],
        first_frame: DataSample,
    ) -> None:
        """
        Open the video container for streaming writes.

        Parameters
        ----------
        sink : Union[str, os.PathLike, BinaryIO]
            Output destination for the video
        first_frame : DataSample
            First frame to determine video properties (resolution, bitrate)

        Raises
        ------
        RuntimeError
            If container is already open
        """
        if self._state.is_open:
            raise RuntimeError("VideoWriter is already open. Call close() first.")

        # Prepare dimensions and bitrate from first frame
        self._state.width, self._state.height = self._prepare(first_frame)

        # Open sink and container with proper cleanup on error
        opened_sink = None
        try:
            if isinstance(sink, (str, Path)):
                opened_sink = open(sink, "wb")
                self._state.sink = opened_sink
                self._state.close_sink = True
            else:
                self._state.sink = sink
                self._state.close_sink = False

            # Open container
            self._state.container = av.open(self._state.sink, mode="w", format=self.extension)

            # Set metadata from first frame's recorded_at if available
            if hasattr(first_frame, "recorded_at") and first_frame.recorded_at is not None:
                first_frame_time = first_frame.recorded_at
                # ISO standard creation time (container level)
                self._state.container.metadata["creation_time"] = first_frame_time.isoformat()
                # QuickTime standard creation date
                self._state.container.metadata["©day"] = first_frame_time.isoformat()
                # Custom field for application-specific queries
                self._state.container.metadata["recorded_at"] = first_frame_time.isoformat()

            # Add stream
            self._state.stream = self._state.container.add_stream(
                self.settings.codec, rate=Fraction(self.frame_rate)
            )
            self._state.stream.width, self._state.stream.height = self._state.width, self._state.height
            self._state.stream.pix_fmt = self.settings.pix_fmt
            self._state.stream.options.update(
                {
                    "crf": str(self.settings.crf),
                    "preset": self.settings.preset,
                    "scenecut": self.settings.scenecut,
                }
            )
            if self.bit_rate:
                self._state.stream.bit_rate = self.bit_rate

            logger.debug(
                f"video writer: {self._state.stream.width}x{self._state.stream.height}, bitrate: {self._state.stream.bit_rate}, pix_fmt: {self._state.stream.pix_fmt}, options: {self._state.stream.options}"
            )
            # Set metadata at track level too
            if hasattr(first_frame, "recorded_at") and first_frame.recorded_at is not None:
                self._state.stream.metadata["creation_time"] = first_frame.recorded_at.isoformat()
                self._state.stream.metadata["recorded_at"] = first_frame.recorded_at.isoformat()
                self._state.stream.metadata["©day"] = first_frame.recorded_at.isoformat()

            self._state.is_open = True
        except Exception:
            if opened_sink:
                opened_sink.close()
            # Reset state to ensure the writer can be safely discarded or reused.
            self._state.reset()
            raise

    def write_frame(self, frame: DataSample) -> None:
        """
        Write a single frame to the video stream.

        Parameters
        ----------
        frame : DataSample
            Frame to encode and write

        Raises
        ------
        RuntimeError
            If container is not open. Call open() first.
        TypeError
            If frame content is not a NumPy ndarray
        """
        if not self._state.is_open:
            raise RuntimeError("VideoWriter is not open. Call open() first.")

        arr = frame.to_ndarray()
        if not isinstance(arr, np.ndarray):
            raise TypeError(f"Expected a NumPy ndarray, got {type(arr).__name__}")

        vf = av.VideoFrame.from_ndarray(arr, format="rgb24")
        # fast, C-level reformat for scaling/pad
        if (vf.width, vf.height) != (self._state.width, self._state.height):
            vf = vf.reformat(width=self._state.width, height=self._state.height, format="rgb24")

        # Encode and mux packets
        for packet in self._state.stream.encode(vf):
            self._state.container.mux(packet)

    def close(self) -> None:
        """
        Flush encoder and close the video container.

        Safe to call multiple times - will only close once.
        """
        if not self._state.is_open:
            return

        # Flush encoder
        for packet in self._state.stream.encode():
            self._state.container.mux(packet)

        # Close container
        self._state.container.close()

        # Close sink if we opened it
        if self._state.close_sink and self._state.sink is not None:
            self._state.sink.close()

        # Reset state
        self._state.reset()

    def write(
        self,
        samples: Iterable[DataSample],
        sink: Union[str, os.PathLike, BinaryIO],
    ) -> None:
        """
        Encode all frames into *sink*, streaming scaled frames via C.

        This method now uses the streaming API internally, avoiding the need
        to load all frames into memory at once.
        """
        # Convert to iterator to peek at first frame
        samples_iter = iter(samples)

        try:
            first_frame = next(samples_iter)
        except StopIteration:
            raise ValueError("VideoWriter.write() received no samples")

        # Use streaming API to avoid loading all frames into memory
        self.open(sink, first_frame)
        try:
            # Write first frame
            self.write_frame(first_frame)

            # Write remaining frames
            for frame in samples_iter:
                self.write_frame(frame)
        finally:
            self.close()



from typing import Tuple

from highlighter.agent.capabilities import Capability, StreamEvent
from highlighter.client.base_models.entities import Entities

__all__ = ["MergeEntities"]


class MergeEntities(Capability):

    class InitParameters(Capability.InitParameters):
        pass

    class StreamParameters(InitParameters):
        pass

    def process_frame(self, stream, data_samples, **kwargs) -> Tuple[StreamEvent, dict]:
        merged_entities = Entities()
        for entities in kwargs.values():
            if not isinstance(entities, (Entities, dict)):
                continue

            merged_entities.merge(entities)
        return StreamEvent.OKAY, {"data_samples": data_samples, "entities": merged_entities}



"""LLM Tool Calling Support

This module provides the infrastructure for LLM capabilities to call tools
like creating case messages, querying data, etc.
"""

import json
import logging
from dataclasses import dataclass
from typing import Any, Dict, List, Optional
from uuid import UUID

from highlighter.client.gql_client import HLClient
from highlighter.client.tasks import TaskContext

__all__ = [
    "ToolExecutionContext",
    "LLMTools",
    "TOOL_USAGE_GUIDELINES",
]


@dataclass
class ToolExecutionContext:
    """Context provided to LLM tool functions.

    This contains all the information tools need to execute actions
    like creating case messages, querying data, etc.
    """

    stream: Any  # Current stream object
    logger: logging.Logger  # LLM capability logger
    client: HLClient  # Highlighter API client
    task_context: Optional[TaskContext]  # Current recording session (if active)
    case_id: Optional[str]  # Current case ID (if recording)

    # From stream variables
    stream_id: str
    capability_name: str  # Name of the capability (for logging)
    data_source_uuid: Optional[UUID] = None
    account_uuid: Optional[UUID] = None


class LLMTools:
    """Container for LLM-callable tool functions.

    Each tool is a static method that receives a ToolExecutionContext
    and returns a structured result dictionary.
    """

    @staticmethod
    def create_message_on_case(
        ctx: ToolExecutionContext,
        message: str,
        message_type: str = "info",
        severity: str = "info",
    ) -> Dict[str, Any]:
        """Add a message to the current recording case.

        This tool allows the LLM to annotate the case with observations,
        alerts, or analysis results.

        Args:
            ctx: Tool execution context
            message: Message content to add to the case
            message_type: Type of message (info, alert, observation, analysis, summary)
            severity: Severity level (info, warning, critical)

        Returns:
            Dict with success status and message_id or error:
            {
                "success": bool,
                "message_id": str (if success),
                "error": str (if failure)
            }
        """
        # Check if case is recording
        if not ctx.case_id:
            ctx.logger.warning(
                "create_message_on_case called but no case is recording",
                extra={"stream_id": ctx.stream_id, "capability_name": ctx.capability_name},
            )
            return {
                "success": False,
                "error": "No case is currently recording",
            }

        try:
            from highlighter.client.base_models.base_models import (
                CreateCaseMessagePayload,
            )

            ctx.logger.info(
                f"LLM tool creating case message: case={ctx.case_id}, "
                f"type={message_type}, severity={severity}, message='{message}'",
                extra={"stream_id": ctx.stream_id, "capability_name": ctx.capability_name},
            )

            # Call Highlighter API to create case message
            result = ctx.client.create_case_message(
                return_type=CreateCaseMessagePayload,
                caseId=ctx.case_id,
                content=message,
            )

            if result.errors:
                ctx.logger.error(
                    f"Failed to create case message: {result.errors}",
                    extra={"stream_id": ctx.stream_id, "capability_name": ctx.capability_name},
                )
                return {"success": False, "error": str(result.errors)}

            # Extract message ID from result
            message_id = None
            if result.user_message:
                message_id = result.user_message.id
            elif result.ai_message:
                message_id = result.ai_message.id

            ctx.logger.info(
                f"Successfully created case message: message_id={message_id}",
                extra={"stream_id": ctx.stream_id, "capability_name": ctx.capability_name},
            )

            return {
                "success": True,
                "message_id": message_id,
                "message": message,
            }

        except Exception as e:
            ctx.logger.error(
                f"Tool execution failed: create_message_on_case - {e}",
                extra={"stream_id": ctx.stream_id, "capability_name": ctx.capability_name},
            )
            return {"success": False, "error": str(e)}

    @staticmethod
    def get_tool_schemas() -> List[Dict]:
        """Return OpenAI/Anthropic function calling schemas for all tools.

        Returns:
            List of tool schemas in the format expected by LLM APIs
        """
        return [
            {
                "type": "function",
                "function": {
                    "name": "create_message_on_case",
                    "description": (
                        "Add a message to the currently recording case. Use this when you detect "
                        "significant events, safety violations, unusual behavior, or anything that "
                        "requires human attention. Only create messages for noteworthy events, not "
                        "routine observations."
                    ),
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "message": {
                                "type": "string",
                                "description": "The message content to add to the case. Be specific and concise.",
                            },
                            "message_type": {
                                "type": "string",
                                "enum": ["info", "alert", "observation", "analysis", "summary"],
                                "description": (
                                    "Type of message: "
                                    "'info' for general information, "
                                    "'alert' for events requiring attention, "
                                    "'observation' for noteworthy observations, "
                                    "'analysis' for analytical findings, "
                                    "'summary' for summary of events"
                                ),
                                "default": "info",
                            },
                            "severity": {
                                "type": "string",
                                "enum": ["info", "warning", "critical"],
                                "description": (
                                    "Severity level: "
                                    "'info' for normal events, "
                                    "'warning' for events that may need attention, "
                                    "'critical' for urgent events requiring immediate attention"
                                ),
                                "default": "info",
                            },
                        },
                        "required": ["message"],
                    },
                },
            }
        ]

    @staticmethod
    def get_tool_function(tool_name: str):
        """Get tool function by name.

        Args:
            tool_name: Name of the tool

        Returns:
            Tool function or None if not found
        """
        return getattr(LLMTools, tool_name, None)


# Prompt template fragment for tool usage guidance
TOOL_USAGE_GUIDELINES = """
## Available Actions

You can take the following actions based on your analysis:

- **create_message_on_case**: Add a message to the current case when you detect:
  * Safety violations or hazards
  * Unusual or anomalous behavior
  * Significant events requiring human attention
  * Equipment malfunctions or failures
  * Important observations or findings

Guidelines:
- Only create messages for noteworthy events, not routine observations
- Be specific and concise in your messages
- Use appropriate severity levels (info/warning/critical)
- Use appropriate message types (info/alert/observation/analysis/summary)
- Avoid creating duplicate messages for the same event

If no significant events are detected, provide a brief summary without creating case messages.
"""



import atexit
import copy
import logging
import threading
from typing import Any, Dict, Optional, Sequence, Tuple

import onnxruntime as ort

from highlighter.core.config import HighlighterRuntimeConfig
from highlighter.core.shared_cache import SharedResourceCache

LOGGER = logging.getLogger(__name__)

_ALLOC_LOCK = threading.RLock()
_SESSION_CACHE: SharedResourceCache[ort.InferenceSession] = SharedResourceCache(maxsize=0)
_ALLOC_STATE: Dict[str, Any] = {
    "initialized": False,
    "enabled": False,
    "signature": None,
}


def _freeze(value: Any) -> Any:
    if isinstance(value, dict):
        return tuple((k, _freeze(v)) for k, v in sorted(value.items()))
    if isinstance(value, (list, tuple)):
        return tuple(_freeze(v) for v in value)
    if isinstance(value, set):
        return tuple(sorted(_freeze(v) for v in value))
    try:
        hash(value)
        return value
    except TypeError:
        return repr(value)


def _split_providers_and_options(
    providers: Sequence[Any], provider_options: Optional[Sequence[Any]]
) -> Tuple[Sequence[str], Optional[Sequence[Any]]]:
    if providers and isinstance(providers[0], (list, tuple)) and len(providers[0]) == 2:
        provider_names = [str(p[0]) for p in providers]
        provider_opts = [p[1] for p in providers]
        return provider_names, provider_opts
    return [str(p) for p in providers], provider_options


def _normalize_provider_options(
    providers: Sequence[str], provider_options: Optional[Sequence[Any]]
) -> list[dict[str, Any]]:
    """Normalize provider options to match providers length (ORT requires this)."""
    num_providers = len(providers)
    if num_providers == 0:
        return []

    if provider_options is None:
        opts: list[Any] = [{} for _ in range(num_providers)]
    elif isinstance(provider_options, dict):
        opts = [provider_options]
    else:
        opts = list(provider_options)

    if len(opts) < num_providers:
        opts.extend({} for _ in range(num_providers - len(opts)))
    return [dict(opt or {}) for opt in opts[:num_providers]]


def _normalize_provider_options_for_key(
    provider_names: Sequence[str], provider_options: Optional[Sequence[Any]]
) -> Tuple[Tuple[str, ...], Tuple[Any, ...]]:
    provider_names_tuple = tuple(provider_names) if provider_names else tuple()
    normalized = _normalize_provider_options(provider_names, provider_options)
    normalized_options = tuple(_freeze(opt) for opt in normalized)
    return provider_names_tuple, normalized_options


def get_ort_options() -> Optional[ort.RunOptions]:
    cfg = copy.deepcopy(HighlighterRuntimeConfig.load().onnxruntime)
    if not cfg.arena_shrinkage:
        return None

    run_options = ort.RunOptions()
    try:
        run_options.add_run_config_entry(
            "memory.enable_memory_arena_shrinkage",
            str(cfg.arena_shrinkage),
        )
    except Exception:
        LOGGER.warning("Unable to set arena shrinkage RunOptions entry.")
        return None
    return run_options


def _arena_cfg_signature(cfg) -> Tuple[Any, ...]:
    return (
        cfg.arena_max_mem_bytes,
        cfg.arena_extend_strategy,
        cfg.arena_initial_chunk_bytes,
    )


def _build_arena_cfg(cfg):
    if not hasattr(ort, "OrtArenaCfg"):
        return None
    if (
        cfg.arena_max_mem_bytes is None
        and cfg.arena_initial_chunk_bytes is None
        and cfg.arena_extend_strategy is None
    ):
        return None

    arena_extend_strategy = int(cfg.arena_extend_strategy or 0)
    max_mem = int(cfg.arena_max_mem_bytes or 0)
    initial_chunk = int(cfg.arena_initial_chunk_bytes or 0)
    max_dead_bytes_per_chunk = 0
    try:
        return ort.OrtArenaCfg(max_mem, arena_extend_strategy, initial_chunk, max_dead_bytes_per_chunk)
    except TypeError:
        LOGGER.warning("OrtArenaCfg signature mismatch; skipping arena config.")
        return None


def _ensure_shared_allocator(cfg) -> bool:
    if not cfg.shared_allocator:
        return False

    with _ALLOC_LOCK:
        signature = _arena_cfg_signature(cfg)
        if _ALLOC_STATE["initialized"]:
            if _ALLOC_STATE["enabled"] and signature != _ALLOC_STATE["signature"]:
                LOGGER.warning(
                    "Shared allocator already initialized with a different config; "
                    "ignoring new allocator settings."
                )
            return _ALLOC_STATE["enabled"]

        if not hasattr(ort, "create_and_register_allocator"):
            LOGGER.warning("onnxruntime.create_and_register_allocator not available; skipping.")
            _ALLOC_STATE["initialized"] = True
            _ALLOC_STATE["enabled"] = False
            _ALLOC_STATE["signature"] = signature
            return False

        arena_cfg = _build_arena_cfg(cfg)
        LOGGER.info(
            "Creating shared allocator (max_mem=%s, extend_strategy=%s, initial_chunk=%s)",
            cfg.arena_max_mem_bytes,
            cfg.arena_extend_strategy,
            cfg.arena_initial_chunk_bytes,
        )
        try:
            mem_info = ort.OrtMemoryInfo(
                "Cpu",
                ort.OrtAllocatorType.ORT_ARENA_ALLOCATOR,
                0,
                ort.OrtMemType.DEFAULT,
            )
            ort.create_and_register_allocator(mem_info, arena_cfg)
            _ALLOC_STATE["initialized"] = True
            _ALLOC_STATE["enabled"] = True
            _ALLOC_STATE["signature"] = signature
            return True
        except Exception:
            LOGGER.exception("Failed to create shared allocator; falling back to per-session.")
            _ALLOC_STATE["initialized"] = True
            _ALLOC_STATE["enabled"] = False
            _ALLOC_STATE["signature"] = signature
            return False


def _apply_cuda_provider_options(
    providers: Sequence[str], provider_options: list[dict[str, Any]], cfg
) -> list[dict[str, Any]]:
    """Apply CUDA config settings. Expects pre-normalized provider_options."""
    if cfg.cuda_arena_extend_strategy is None and cfg.cuda_gpu_mem_limit_bytes is None:
        return provider_options

    for idx, provider in enumerate(providers):
        if provider != "CUDAExecutionProvider":
            continue
        opts = dict(provider_options[idx])
        if cfg.cuda_arena_extend_strategy is not None:
            opts["arena_extend_strategy"] = cfg.cuda_arena_extend_strategy
        if cfg.cuda_gpu_mem_limit_bytes is not None:
            opts["gpu_mem_limit"] = int(cfg.cuda_gpu_mem_limit_bytes)
        provider_options[idx] = opts

    return provider_options


def _build_session_options(cfg, custom_ops_path: Optional[str], use_shared_allocator: bool):
    sess_options = ort.SessionOptions()
    if custom_ops_path:
        sess_options.register_custom_ops_library(str(custom_ops_path))

    if hasattr(sess_options, "enable_cpu_mem_arena"):
        sess_options.enable_cpu_mem_arena = bool(cfg.enable_cpu_mem_arena)
    else:
        LOGGER.warning("SessionOptions.enable_cpu_mem_arena not available; skipping.")

    if cfg.use_global_threadpool:
        if hasattr(sess_options, "use_per_session_threads"):
            sess_options.use_per_session_threads = False
        else:
            LOGGER.warning("SessionOptions.use_per_session_threads not available; skipping.")

    if use_shared_allocator:
        try:
            sess_options.add_session_config_entry("session.use_env_allocators", "1")
        except Exception:
            LOGGER.warning("Unable to enable env allocators for shared allocator.")

    return sess_options


def _cache_key(
    model_path: str,
    providers: Sequence[Any],
    provider_options: Optional[Sequence[Any]],
    custom_ops_path: Optional[str],
    cfg,
) -> Tuple[Any, ...]:
    provider_names, provider_opts = _split_providers_and_options(providers, provider_options)
    _, normalized_options = _normalize_provider_options_for_key(provider_names, provider_opts)
    return (
        str(model_path),
        tuple(provider_names),
        normalized_options,
        str(custom_ops_path) if custom_ops_path else None,
        cfg.shared_allocator,
        cfg.enable_cpu_mem_arena,
        cfg.use_global_threadpool,
        cfg.arena_extend_strategy,
        cfg.arena_max_mem_bytes,
        cfg.arena_initial_chunk_bytes,
        cfg.cuda_arena_extend_strategy,
        cfg.cuda_gpu_mem_limit_bytes,
    )


def get_session(
    model_path: str,
    *,
    providers: Sequence[Any],
    provider_options: Optional[Sequence[Any]] = None,
    custom_ops_path: Optional[str] = None,
) -> ort.InferenceSession:
    cfg = copy.deepcopy(HighlighterRuntimeConfig.load().onnxruntime)

    key = _cache_key(model_path, providers, provider_options, custom_ops_path, cfg)
    provider_names_list: list[str] = []  # Populated only by the creator thread for logging.
    shared_allocator_enabled = False

    def _create_session() -> ort.InferenceSession:
        nonlocal shared_allocator_enabled, provider_names_list
        shared_allocator_enabled = _ensure_shared_allocator(cfg)
        sess_options = _build_session_options(cfg, custom_ops_path, shared_allocator_enabled)

        provider_names_list, provider_options_raw = _split_providers_and_options(providers, provider_options)
        provider_options_list = _normalize_provider_options(provider_names_list, provider_options_raw)
        provider_options_list = _apply_cuda_provider_options(provider_names_list, provider_options_list, cfg)

        return ort.InferenceSession(
            str(model_path),
            sess_options=sess_options,
            providers=provider_names_list,
            provider_options=provider_options_list,
        )

    try:
        session = _SESSION_CACHE.get_or_create(
            key,
            _create_session,
            maxsize=int(cfg.session_cache_maxsize),
        )
    except Exception:
        LOGGER.exception(
            "Failed to create InferenceSession for model_path=%s providers=%s",
            model_path,
            providers,
        )
        raise

    if provider_names_list:
        LOGGER.debug(
            "Created ORT session (model=%s, providers=%s, allocator=%s)",
            model_path,
            provider_names_list,
            "shared" if shared_allocator_enabled else "per-session",
        )
    return session


def clear_cache() -> None:
    _SESSION_CACHE.clear()


atexit.register(clear_cache)



# Declaration order is based on the dependency on static references
#
# To Do
# ~~~~~
# - None, yet !

from .message import Message, MessageState

from .castaway import Castaway

from .mqtt import MQTT



# Aiko Services: Storage SPI
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# To Do
# ~~~~~
# - Complete Interface method signatures using "from typing"
#
# - Consider "dependency_name" --> "dependency_path"
# - Consider "category_name" --> "category_path"
# - Consider "entry_name" --> "entry_path"

from abc import abstractmethod

from aiko_services.main import Actor, Interface

__all__ = ["Storage"]

# --------------------------------------------------------------------------- #

class Storage(Actor):
    Interface.default("Storage",
        "aiko_services.main.storage.storage_file.StorageFileImpl")

    @abstractmethod
    def add(self, dependency_name, dependency=None):
        pass

    @abstractmethod
    def create(self, category_name):
        pass

    @abstractmethod
    def destroy(self, entry_name):     # Entry is either Category or Dependency
        pass

    @abstractmethod
    def dump(self, sort_by_name):
        pass

    @abstractmethod
    def exit(self):
        pass

    @abstractmethod
    def initialize(self, storage_url):
        pass

    @abstractmethod
    def link(self, entry_path_new, entry_path_existing):
        pass

    @abstractmethod
    def list(self, topic_path_response, entry_name=None,
        long_format=False, recursive=False, entry_records=None):
        pass

    @abstractmethod
    def remove(self, entry_name):      # Entry is either Category or Dependency
        pass

    @abstractmethod
    def update(self, entry_name, service=None,
        service_filter=None, lifecycle_manager_url=None, storage_url=None):
        pass

# --------------------------------------------------------------------------- #



try:
    import ultralytics
except ModuleNotFoundError as _:
    raise ModuleNotFoundError(
        "The ultralytics package is not installed, use " "`pip install highlighter-sdk[yolo]`"
    )

__all__ = ["generate", "train", "prepare_datasets"]
from .trainer import YoloV11Trainer



"""Data source management package for Highlighter."""

from highlighter.datasource.models import (
    DataSourceTemplate,
    LocalDataSourceEntry,
    LocalDataSourceFile,
    SyncResult,
)
from highlighter.datasource.service import DataSourceService

__all__ = [
    "DataSourceService",
    "LocalDataSourceEntry",
    "LocalDataSourceFile",
    "DataSourceTemplate",
    "SyncResult",
]



"""
Deprecated: Image models are deprecated in favor of File models.
Use File, FilePresigned, and FileConnection instead.
"""

from .file import File, FileConnection, FilePresigned

__all__ = ["Image", "ImagePresigned", "ImageConnection"]

# Backwards compatibility aliases
Image = File
ImagePresigned = FilePresigned
ImageConnection = FileConnection



import logging
import os
import re
import sys
import tempfile
from logging.handlers import RotatingFileHandler
from typing import Any, Mapping, Optional

from aiko_services.main.utilities.logger import (
    _LOG_FORMAT_DATETIME,
    _LOG_FORMAT_DEFAULT,
)

__all__ = [
    "configure_root_logger",
    "TRACE",
    "VERBOSE",
    "COLOURS",
    "format_stream_info",
    "format_timing",
    "safe_qsize",
    "log_queue_size",
]


# ----- Define extra levels -----
TRACE = 5
VERBOSE = 8

logging.addLevelName(TRACE, "TRACE")
logging.addLevelName(VERBOSE, "VERBOSE")


# Add convenience methods to Logger class
def _trace(self, message, *args, **kwargs):
    if self.isEnabledFor(TRACE):
        self._log(TRACE, message, args, **kwargs)


def _verbose(self, message, *args, **kwargs):
    if self.isEnabledFor(VERBOSE):
        self._log(VERBOSE, message, args, **kwargs)


logging.Logger.trace = _trace
logging.Logger.verbose = _verbose

# Terse log format - compact and easy to read
# Format: MM-DD HH:MM:SS [LVL] module:line message
TERSE_LOG_FORMAT = "%(asctime)s [%(levelname).3s] %(message)s   |> %(name)s#%(funcName)s:%(lineno)d"

# More verbose debug format for deep debugging
DEBUG_LOG_FORMAT = "%(asctime)s.%(msecs)03d %(levelname)-8s %(threadName)s %(message)s   |> %(name)s#%(funcName)s:%(lineno)-6d"

RESET = "\033[0m"

# Named colors for manual use via extra={"color": "red"}
COLOURS = {
    "grey": "\033[90m",
    "light_grey": "\033[37m",
    "red": "\033[31m",
    "green": "\033[32m",
    "yellow": "\033[33m",
    "blue": "\033[34m",
    "magenta": "\033[35m",
    "cyan": "\033[36m",
    "white": "\033[97m",
    "bold": "\033[1m",
}

# Color map for each log level
# INFO uses standard color, lower levels use dimmer colors, higher levels use bright colors
COLOUR_MAP = {
    TRACE: "\033[90m",  # grey (dimmest)
    VERBOSE: "\033[37m",  # light grey
    logging.DEBUG: "\033[36m",  # cyan (light)
    logging.INFO: None,  # standard/no color
    logging.WARNING: "\033[33m",  # yellow
    logging.ERROR: "\033[31m",  # red
    logging.CRITICAL: "\033[41m",  # red background
}


class ColourFormatter(logging.Formatter):
    """Formatter that adds colors based on log level or custom color attribute.

    Supports multiple modes:
    1. Automatic level-based coloring (default)
    2. Manual color override via extra parameter:
       logger.info("message", extra={"color": "red"})
    3. Stream info formatting via extra parameters:
       logger.info("message", extra={"stream_id": "7", "capability_name": "VideoDataSource"})
    """

    def format(self, record):
        # Add stream info prefix if provided via extra
        stream_id = getattr(record, "stream_id", None)
        capability_name = getattr(record, "capability_name", None)
        frame_id = getattr(record, "frame_id", None)

        if stream_id is not None:
            stream_prefix = format_stream_info(stream_id, capability_name, frame_id)
            # Store original message and prepend stream info
            original_msg = record.getMessage()
            record.msg = f"{stream_prefix}: {original_msg}"
            record.args = ()  # Clear args since we've already formatted the message

        msg = super().format(record)

        # Color the tail (module.function:line) in grey
        tail = f"|> {record.name}#{record.funcName}:{record.lineno}"
        coloured_tail = f"{COLOURS['grey']}{tail}{RESET}"
        msg = msg.replace(tail, coloured_tail)

        # Check for manual color override first
        color = getattr(record, "color", None)
        if color and color in COLOURS:
            return f"{COLOURS[color]}{msg}{RESET}"

        # Fall back to level-based colors for console output
        color_code = COLOUR_MAP.get(record.levelno)
        if color_code:
            return f"{color_code}{msg}{RESET}"

        return msg


class TerseNameFilter(logging.Filter):
    """Remove redundant 'highlighter.' prefix from logger names."""

    def filter(self, record):
        if record.name.startswith("highlighter."):
            record.name = record.name[12:]  # Remove "highlighter." prefix
        return True


def is_running_under_pytest():
    """Check if we're actually running under pytest (not just if pytest is installed)"""
    # More robust check: look for PYTEST_CURRENT_TEST environment variable
    # which is only set when actually running tests
    return os.getenv("PYTEST_CURRENT_TEST") is not None


def configure_root_logger(
    _log_path: Optional[str] = None,
    _log_level: Optional[str] = None,
    _log_rotation_max_kilobytes: Optional[int] = None,
    _log_rotation_backup_count: Optional[int] = None,
):
    """Configure the root logger with file and stream handlers.

    Supports HL_LOG_LEVEL environment variable:
    - HL_LOG_LEVEL: Set specific log level (TRACE, VERBOSE, DEBUG, INFO, WARNING, ERROR, CRITICAL)

    Args:
        _log_path: Optional path for log file
        _log_level: Optional explicit log level (overrides environment variable)
        _log_rotation_max_kilobytes: Optional max size for log rotation
        _log_rotation_backup_count: Optional number of backup files
    """
    if _log_path is None:
        temp_file = tempfile.NamedTemporaryFile(suffix=".log", delete=False)
        _log_path = temp_file.name
        temp_file.close()

    # Only check environment variables if _log_level wasn't explicitly passed
    if _log_level is None:
        _log_level = os.getenv("HL_LOG_LEVEL", "WARNING")

    # Set defaults for log rotation parameters
    if _log_rotation_max_kilobytes is None:
        _log_rotation_max_kilobytes = 100 * 1024  # 100 MB in KB
    if _log_rotation_backup_count is None:
        _log_rotation_backup_count = 4

    root = logging.getLogger()
    if root.hasHandlers():
        for handler in root.handlers:
            handler.close()
        root.handlers.clear()

    # Use terse format by default, DEBUG format only for DEBUG level
    log_format = DEBUG_LOG_FORMAT if _log_level == "DEBUG" else TERSE_LOG_FORMAT
    datefmt = "%m-%d %H:%M:%S" if _log_level != "DEBUG" else _LOG_FORMAT_DATETIME

    # Use colored formatter for console output
    color_formatter = ColourFormatter(log_format, datefmt=datefmt)

    # Use plain formatter for file output
    plain_formatter = logging.Formatter(log_format, datefmt=datefmt)

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(color_formatter)
    stream_handler.addFilter(TerseNameFilter())  # Strip "highlighter." prefix

    handlers = [stream_handler]

    log_level_mapping = {
        "TRACE": TRACE,
        "VERBOSE": VERBOSE,
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }
    log_level = log_level_mapping.get(_log_level)
    if log_level is None:
        raise SystemExit(f"Invalid log_level '{_log_level}'")

    if not is_running_under_pytest():
        # Ensure log_path exists
        directory = os.path.dirname(_log_path)
        os.makedirs(directory, exist_ok=True)

        if not os.path.exists(_log_path):
            with open(_log_path, "w") as file:
                file.write("")  # Creates an empty file

        # Setup File Handler
        file_handler = RotatingFileHandler(
            _log_path,
            maxBytes=_log_rotation_max_kilobytes * 1024,  # Convert KB to bytes
            backupCount=_log_rotation_backup_count,
        )
        file_handler.setFormatter(plain_formatter)
        file_handler.addFilter(TerseNameFilter())  # Strip "highlighter." prefix

        handlers.append(file_handler)
    logging.basicConfig(handlers=handlers)
    # Set the log level for highlighter code
    logging.getLogger("highlighter").setLevel(log_level)
    logging.getLogger(__name__).info(f"log_path: {_log_path}")

    for module_path, module_log_level in get_log_level_env_vars().items():
        logging.getLogger(module_path).setLevel(module_log_level)
        logging.getLogger(__name__).info(f"Set log level for {module_path} to {module_log_level}")


def get_log_level_env_vars():
    """Lookup log-level directives from environment variables
    matching the pattern:
        LOG_LEVEL_module_DOT_path_DOT_segment=INFO
    into {"module.path.segment": "INFO"}
    """
    pattern = re.compile(r"^LOG_LEVEL_(.*)$")
    results = {}
    for key, value in os.environ.items():
        match = pattern.match(key)
        if match:
            # Replace all "_DOT_" with "." in the matched part
            dotted_path = match.group(1).replace("_DOT_", ".")
            results[dotted_path] = value
    return results


class ColourStr:
    HEADER = "\033[95m"

    BLUE = "\033[94m"
    CYAN = "\033[96m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"

    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

    RESET = "\033[0m"

    @staticmethod
    def blue(s):
        return ColourStr.BLUE + s + ColourStr.RESET

    @staticmethod
    def cyan(s):
        return ColourStr.CYAN + s + ColourStr.RESET

    @staticmethod
    def green(s):
        return ColourStr.GREEN + s + ColourStr.RESET

    @staticmethod
    def red(s):
        return ColourStr.RED + s + ColourStr.RESET

    @staticmethod
    def yellow(s):
        return ColourStr.YELLOW + s + ColourStr.RESET

    @staticmethod
    def bold(s):
        return ColourStr.BOLD + s + ColourStr.RESET

    @staticmethod
    def underline(s):
        return ColourStr.UNDERLINE + s + ColourStr.RESET


def format_stream_info(stream_id, capability_name=None, frame_id=None):
    """Format stream information with consistent colors.

    Args:
        stream_id: Stream identifier (string or number)
        capability_name: Optional capability name
        frame_id: Optional frame identifier

    Returns:
        Formatted string with colors: "<Stream <id>:<frame_id>| Capability>"

    Example:
        format_stream_info("7", "VideoDataSource", "23")
        # Returns colored: "<Stream 7:23| VideoDataSource>"
    """
    stream_id_str = str(stream_id) if stream_id is not None else "unknown"

    # Add frame_id if provided
    if frame_id is not None:
        stream_id_str = f"{stream_id_str}:{frame_id}"

    # Color stream ID in cyan
    colored_stream = f"<{COLOURS['cyan']}Stream {stream_id_str}{RESET}|"

    if capability_name:
        # Color capability name in magenta
        colored_cap = f"{COLOURS['magenta']}{capability_name}{RESET}>"
        return f"{colored_stream} {colored_cap}"

    return colored_stream


def format_timing(label, seconds, unit="s", color_threshold=None):
    """Format timing information with consistent colors.

    Args:
        label: Label for the timing (e.g., "S3", "upload", "save_local")
        seconds: Time in seconds
        unit: Unit to display (default: "s" for seconds)
        color_threshold: Optional dict with thresholds {"fast": 1.0, "slow": 5.0}
                        Values under "fast" are green, over "slow" are yellow

    Returns:
        Formatted string with colors: "label: X.XXs"

    Example:
        format_timing("upload", 2.5, color_threshold={"fast": 1.0, "slow": 5.0})
        # Returns colored: "upload: 2.5s" (in yellow if >5s, green if <1s, default otherwise)
    """
    # Choose color based on threshold if provided
    if color_threshold:
        if seconds < color_threshold.get("fast", float("inf")):
            color = COLOURS["green"]
        elif seconds > color_threshold.get("slow", float("inf")):
            color = COLOURS["yellow"]
        else:
            color = ""
    else:
        color = ""

    # Format the timing value
    if unit == "s":
        value_str = f"{seconds:.3f}{unit}"
    else:
        value_str = f"{seconds:.1f}{unit}"

    if color:
        return f"{label}: {color}{value_str}{RESET}"
    else:
        return f"{label}: {value_str}"


def safe_qsize(
    queue,
    logger: Optional[logging.Logger] = None,
    *,
    log_level: int = logging.DEBUG,
    log_message: str = "Unable to read queue size",
    extra: Optional[dict] = None,
    stacklevel: int = 2,
) -> Optional[int]:
    """Best-effort queue size retrieval with optional debug logging."""
    if queue is None:
        if logger and logger.isEnabledFor(log_level):
            logger.log(
                log_level,
                f"{log_message} (queue is None)",
                extra=extra,
                stacklevel=stacklevel,
            )
        return None

    qsize_fn = getattr(queue, "qsize", None)
    if qsize_fn is None:
        if logger and logger.isEnabledFor(log_level):
            logger.log(
                log_level,
                f"{log_message} (missing qsize)",
                extra=extra,
                stacklevel=stacklevel,
            )
        return None

    try:
        return qsize_fn()
    except Exception as exc:
        if logger and logger.isEnabledFor(log_level):
            logger.log(log_level, log_message, exc_info=exc, extra=extra, stacklevel=stacklevel)
        return None


def log_queue_size(
    logger: logging.Logger,
    level: int,
    queue,
    message: str,
    *,
    message_args: Optional[Mapping[str, Any]] = None,
    min_size: Optional[int] = None,
    error_message: str = "Unable to read queue size",
    **kwargs,
) -> Optional[int]:
    """Log queue size as optional telemetry without littering call sites.

    Queue size is best-effort: some queue types/platforms raise or don't implement
    qsize(), and callers generally don't want those failures to interrupt control
    flow. This helper centralizes that behavior and provides consistent debug
    logging when qsize can't be read.

    The queue size is injected into `message_args` as "qsize", so `message` should
    include a mapping placeholder (e.g. "Queue size: %(qsize)s").
    """
    if not logger.isEnabledFor(level):
        return None

    log_kwargs = dict(kwargs)
    stacklevel = log_kwargs.pop("stacklevel", 2)
    extra = log_kwargs.get("extra")

    qsize = safe_qsize(
        queue,
        logger=logger,
        log_message=error_message,
        extra=extra,
        stacklevel=stacklevel + 1,
    )
    if qsize is None:
        return None

    if min_size is not None and qsize < min_size:
        return qsize

    log_kwargs["stacklevel"] = stacklevel
    args = dict(message_args or {})
    args["qsize"] = qsize
    try:
        formatted_message = message % args
    except Exception as exc:
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                "Unable to format queue-size log message",
                exc_info=exc,
                extra=extra,
                stacklevel=stacklevel,
            )
        return qsize
    logger.log(level, formatted_message, **log_kwargs)
    return qsize



import logging
from dataclasses import field
from datetime import datetime, timezone
from typing import List, Optional, Tuple, Union
from uuid import UUID

import numpy as np
from PIL import Image as PILImage
from pydantic import BaseModel, ConfigDict
from shapely.geometry import Polygon

from highlighter.client.base_models.entities import Entities
from highlighter.client.base_models.observation import Observation
from highlighter.core import draw_annotations_on_image
from highlighter.core.enums import ContentTypeEnum

from .. import resize_image

logger = logging.getLogger(__name__)


class DataSample(BaseModel):
    """
    Video / image sample
    """

    content: Optional[bytes | str | np.ndarray | PILImage.Image | Observation | Entities] = None
    content_type: ContentTypeEnum = ContentTypeEnum.UNKNOWN
    recorded_at: datetime = field(default_factory=lambda: datetime.now(timezone.utc))  # FIXME remove default
    stream_frame_index: int = 0
    media_frame_index: int = 0
    # TODO: Replace with data_source_id once we re-target annotations and submissions from data_files to data_sources
    data_file_id: Optional[UUID] = None

    # Configure pydantic to allow np.ndarray types
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __post_init__(self):
        if not isinstance(self.content, (bytes, str, np.ndarray, PILImage.Image, Observation)):
            raise ValueError(
                f"Content must be bytes, str, np.ndarray, PIL.Image or Observation, got {type(self.content)}"
            )

    # TODO: handle non image data samples
    def to_ndarray(self) -> np.ndarray:
        """Always return an ``numpy.ndarray`` (H, W, 3, uint8)."""
        if isinstance(self.content, np.ndarray):
            return self.content
        if isinstance(self.content, PILImage.Image):
            return np.asarray(self.content)
        raise TypeError("Unsupported sample type: expected ndarray or PIL.Image.")

    @property
    def wh(self) -> Tuple[int, int]:
        assert self.content_type == "image"
        if isinstance(self.content, np.ndarray):
            h, w = self.content.shape[:2]
        elif isinstance(self.content, PILImage.Image):
            w, h = self.content.size
        else:
            raise ValueError(f"Invalid image type, got {type(self.content)}")
        return w, h

    def resize_content(self, width: int, height: int) -> Union[np.ndarray, PILImage.Image]:
        assert self.content_type == "image"
        return resize_image(self.content, width=width, height=height)

    def crop_content(
        self,
        locations: Union[
            List[Polygon], List[Tuple[int, int, int, int]], List[Tuple[float, float, float, float]]
        ],
        crop_args: Optional["CropArgs"] = None,
        as_data_sample: bool = False,
    ) -> List[Union[np.ndarray, PILImage.Image]] | List["DataSample"]:
        from highlighter.datasets.cropping import crop_rect_from_poly

        assert self.content_type == "image"
        if not locations:
            return []

        # If proportional crop locations convert to absolute for cropping
        if isinstance(locations[0], tuple) and isinstance(locations[0][0], float):
            ori_w, ori_h = self.wh
            polys = [
                Polygon(
                    [
                        (int(x0 * ori_w), int(y0 * ori_h)),
                        (int(x1 * ori_w), int(y0 * ori_h)),
                        (int(x1 * ori_w), int(y1 * ori_h)),
                        (int(x0 * ori_w), int(y1 * ori_h)),
                        (int(x0 * ori_w), int(y0 * ori_h)),
                    ]
                )
                for x0, y0, x1, y1 in locations
            ]

        elif isinstance(locations[0], tuple) and isinstance(locations[0][0], int):
            ori_w, ori_h = self.wh
            polys = [
                Polygon(
                    [
                        (x0, y0),
                        (x1, y0),
                        (x1, y1),
                        (x0, y1),
                        (x0, y0),
                    ]
                )
                for x0, y0, x1, y1 in locations
            ]
        elif isinstance(locations[0], Polygon):
            polys = locations
        else:
            raise ValueError(f"Invalid crop locations, got, {locations}.")

        crop_contents = crop_rect_from_poly(self.content, polys, crop_args)
        if as_data_sample:
            return [
                DataSample(
                    content=c,
                    content_type=self.content_type,
                    recorded_at=self.recorded_at,
                    stream_frame_index=self.stream_frame_index,
                    media_frame_index=self.media_frame_index,
                    data_file_id=self.data_file_id,
                )
                for c in crop_contents
            ]
        else:
            return crop_contents

    def draw_annotations(self, annotations: "Annotations"):
        overlay = draw_annotations_on_image(self.content, annotations)
        return overlay



"""CUDA availability helpers shared across predictors.

These helpers cache results per process (via functools.lru_cache) to avoid
repeated subprocess calls during model initialization.
"""

import functools
import os
import subprocess  # nosec B404 - static nvidia-smi call used for availability check


@functools.lru_cache(maxsize=1)
def is_nvidia_smi_available() -> bool:
    """Return True if nvidia-smi is available and responds successfully."""
    try:
        # [B603:subprocess_without_shell_equals_true] subprocess call - check for execution of untrusted input.
        # [B607:start_process_with_partial_path] Starting a process with a partial executable path
        subprocess.run(
            ["nvidia-smi"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )  # nosec B603, B607
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False


def resolve_device_id(env_var: str = "HL_GPU_DEVICE_ID", default: int = 0) -> int:
    """Return an integer device id from an env var, falling back to default on errors."""
    raw = os.environ.get(env_var, default)
    try:
        return int(raw)
    except (TypeError, ValueError):
        return int(default)



# Usage
# ~~~~~
# from aiko_services.main.utilities import *
# lock_test = Lock(f"{__name__}.test", aiko.logger(__name__))
# try:
#     lock_test.acquire("add_service()")
#     ## Critial code ##
# finally:
#     lock_test.release()
#
# Notes
# ~~~~~
# - Generally useful debugging aid: export AIKO_LOG_LEVEL_LOCK=DEBUG
#
# To Do
# ~~~~~
# - Consider keeping a process wide list of all the Locks, which can be shown
#
# - Replace all Locks (not time critical), by ensuring all updates occur via
#   "events" (handlers, messages and timers) managed by the event loop and
#   executed solely by the event loop thread

import os
from threading import Lock as ThreadingLock

from .logger import DEBUG, get_logger

_AIKO_LOG_LEVEL_LOCK = os.environ.get("AIKO_LOG_LEVEL_LOCK", "INFO")
_LOGGER = get_logger(__name__, log_level=_AIKO_LOG_LEVEL_LOCK)

__all__ = ["Lock"]

class Lock:
    def __init__(self, name, logger=None):
        self._name = name
        self._logger = logger
        self._lock = ThreadingLock()
        self._in_use = None

    def acquire(self, location):
        if self._in_use and _LOGGER.isEnabledFor(DEBUG):
            _LOGGER.debug(
                f'"{self._name}" at "{location}" in use by "{self._in_use}"')

        self._lock.acquire()
        self._in_use = location

        if _LOGGER.isEnabledFor(DEBUG):
            _LOGGER.debug(f'"{self._name}" acquired by {location}')

    def in_use(self):
        return self._in_use

    def release(self):
        if _LOGGER.isEnabledFor(DEBUG):
            _LOGGER.debug(f'"{self._name}" released by {self._in_use}')

        self._in_use = None
        self._lock.release()



# To Do
# ~~~~~
# - Implement disconnect() and support for "with" statement ?
# - Implement statistics measurements, e.g performance

import abc
from typing import Any

__all__ = ["Message", "MessageState"]

class MessageState:
    DISCONNECTED = "DISCONNECTED"
    CONNECTED = "CONNECTED"

class Message(abc.ABC):
    def __init__(
        self: Any,
        message_handler: Any = None,
        topics_subscribe: Any = None,
        topic_lwt: str = None,
        payload_lwt: str = None,
        retain_lwt: bool = False,
        mqtt_state_handler = None
        ) -> None:

        pass

    def publish(
        self: Any,
        topic: str,
        payload: Any,
        retain: bool = False,
        wait: bool = False
        ) -> None:

        raise NotImplementedError("Message.publish()")

    def set_last_will_and_testament(
        self: Any,
        topic_lwt: str = None,
        payload_lwt: str = "(absent)",
        retain_lwt: bool = False
        ) -> None:

        raise NotImplementedError("Message.set_last_will_and_testament()")

    def subscribe(self: Any, topics: Any) -> None:
        raise NotImplementedError("Message.subscribe()")

    def unsubscribe(self: Any, topics: Any, remove=True) -> None:
        raise NotImplementedError("Message.unsubscribe()")



from aiko_services.main.utilities import importer

__all__ = ["patch_capability"]


def patch_capability(module_import_str, patch_attr, patch):
    module = importer.load_module(module_import_str)
    setattr(module, patch_attr, patch)



"""Agent-level recorder management.

This module provides the AgentRecorderManager class which manages the lifecycle
of recorders for agent-level observation recording. Each stream gets a single
recorder that is shared by all capabilities in that stream.
"""

import logging
import threading
from typing import Dict, List, Optional

from sqlalchemy.orm import Session

from highlighter.agent.capabilities.base_capability import _AppendableIterator
from highlighter.agent.capabilities.recorder import Recorder, RecordMode
from highlighter.cli.cli_logging import safe_qsize
from highlighter.core.enums import ContentTypeEnum


class AgentRecorderManager:
    """Manages agent-level recorders for observation recording.

    Handles the lifecycle of recorders that are shared across capabilities in a stream.
    Each stream gets a single recorder that all capabilities in that stream write to.

    This separates recorder management concerns from HLAgent, making the code more
    testable and maintainable.
    """

    def __init__(self, pipeline, logger: Optional[logging.Logger] = None):
        """Initialize the recorder manager.

        Args:
            pipeline: The pipeline instance to scan for capabilities
            logger: Optional logger instance (creates one if not provided)
        """
        self.pipeline = pipeline
        self.logger = logger or logging.getLogger(__name__)

        # Recorder state
        self._recorders: Dict[str, Recorder] = {}  # stream_id -> Recorder
        # stream_id -> sample queue feeding the Recorder (push from capabilities, pull by Recorder)
        self._data_sample_iterators: Dict[str, _AppendableIterator] = {}
        self._recorder_init_lock = threading.Lock()  # Protect lazy initialization
        self._initialized_streams = set()  # Track which streams have initialized recorders
        self._stream_graph_paths: Dict[str, str] = {}  # stream_id -> graph_path

    def register_stream(self, stream_id: str, graph_path: str = None):
        """Register a stream's graph path for later use during recorder initialization.

        Args:
            stream_id: The stream ID to register
            graph_path: The graph path for this stream (defaults to main graph if None)
        """
        # If no graph_path provided, use the main graph (first element in pipeline graph)
        if graph_path is None:
            # Get the default graph path from pipeline graph
            graph_items = list(self.pipeline.pipeline_graph._graph.items())
            if graph_items:
                graph_path = graph_items[0][0]  # First graph node name
            else:
                graph_path = "default"  # Fallback

        self._stream_graph_paths[stream_id] = graph_path

    def ensure_initialized(self, stream_id: str):
        """Lazily initialize recorders for a stream on first use.

        This must be called from a context where the stream's thread-local storage
        is set up (e.g., during frame processing), not during create_stream().

        Thread-safe: uses a lock to prevent concurrent initialization of the same stream.

        Args:
            stream_id: The stream ID to initialize recorders for
        """
        # Fast path: check without lock if already initialized
        if stream_id in self._initialized_streams:
            return

        # Slow path: acquire lock and initialize if needed
        with self._recorder_init_lock:
            # Double-check after acquiring lock (another thread might have initialized)
            if stream_id in self._initialized_streams:
                return

            # Now safe to initialize since we're in a context with thread-local stream access
            self._initialize_for_stream(stream_id)
            self._initialized_streams.add(stream_id)

    def _initialize_for_stream(self, stream_id: str):
        """Initialize recorder for a stream.

        Creates a single recorder per stream that is shared by all capabilities in that stream.

        Args:
            stream_id: The stream ID to initialize recorder for
        """
        # Find any capability with recording enabled to get recording parameters
        recording_params = None
        recording_capabilities = []

        # Get the graph path from our stored mapping, fallback to default behavior
        head_node_name = self._stream_graph_paths.get(stream_id)
        if head_node_name is None:
            # Fallback: try to get from thread-local context if available
            try:
                head_node_name = self.pipeline.get_stream()[0].graph_path
            except AttributeError:
                # If thread-local access fails, use the first graph node as default
                graph_items = list(self.pipeline.pipeline_graph._graph.items())
                head_node_name = graph_items[0][0] if graph_items else "default"
                self.logger.warning(
                    f"Stream {stream_id} graph_path not registered, using default: {head_node_name}"
                )

        graph_path = self.pipeline.pipeline_graph.get_path(head_node_name)

        for node in graph_path:
            node_name = node.name
            capability = node.element

            # Check if this capability supports recording
            if not hasattr(capability, "recording_enabled"):
                continue

            # Check if recording is enabled for this capability and stream
            if not capability.recording_enabled(stream_id):
                continue

            recording_capabilities.append(node_name)

            # Get parameters from the first capability with recording enabled
            if recording_params is None:
                recording_params = capability.stream_parameters(stream_id)

        # If no capabilities have recording enabled, nothing to do
        if not recording_params:
            return

        database = getattr(recording_params, "database", None)
        record = getattr(recording_params, "record", None) or RecordMode.OFF

        # Only create recorder if record mode is not OFF
        if record is not RecordMode.OFF:
            # Validate required parameters before creating recorder
            if database is None:
                raise ValueError(f"Missing 'database' for stream '{stream_id}', required when recording")

            self.logger.info(
                f"Recording {len(recording_capabilities)} capabilities "
                f"({', '.join(recording_capabilities)}) to {record}",
                extra={"stream_id": stream_id, "capability_name": "AgentOutputRecorder"},
            )

            # Create appendable iterator for this stream
            data_sample_iterator = _AppendableIterator()
            self._data_sample_iterators[stream_id] = data_sample_iterator

            # Create recorder for this stream
            self._recorders[stream_id] = Recorder(
                iterator=data_sample_iterator,
                session_factory=lambda: Session(database.engine),
                content_type=ContentTypeEnum.ENTITIES,
                account_uuid=recording_params.data_source_uuid,
                data_source_uuid=recording_params.data_source_uuid,
                recording_state=recording_params.recording_state,
                record=recording_params.record,
                samples_per_file=recording_params.samples_per_file,
                seconds_per_file=recording_params.seconds_per_file,
                output_filename_template=recording_params.output_filename_template,
                output_folder=recording_params.output_folder,
                stream_id=stream_id,
                capability_name="AgentOutputRecorder",
            )

    def cleanup_stream(self, stream_id: str):
        """Cleanup recorders for a stream.

        Flushes and removes the recorder for this stream.

        Args:
            stream_id: The stream ID to cleanup
        """
        # Clean up this stream's recorder
        if stream_id in self._recorders:
            self._recorders[stream_id].flush()  # Blocks until recorder worker thread exits
            del self._recorders[stream_id]
        if stream_id in self._data_sample_iterators:
            del self._data_sample_iterators[stream_id]

        # Remove stream from initialized set and graph path mapping
        self._initialized_streams.discard(stream_id)
        self._stream_graph_paths.pop(stream_id, None)

    def record_data_sample(self, stream_id: str, data_sample):
        """Record a data sample to the stream's recorder.

        Args:
            stream_id: The stream ID to record to
            data_sample: The data sample to record
        """
        if stream_id not in self._recorders:
            return
        iterator = self._data_sample_iterators.get(stream_id)
        if iterator is None:
            return

        iterator.append(data_sample)
        if self.logger.isEnabledFor(logging.INFO):
            extra = {"stream_id": stream_id, "capability_name": "AgentOutputRecorder"}
            qsize = safe_qsize(
                iterator,
                logger=self.logger,
                log_message="Unable to read recorder queue size",
                extra=extra,
            )
            if qsize is not None:
                maxsize = iterator.maxsize()
                if maxsize > 0 and qsize >= max(1, maxsize - 2):
                    self.logger.info(
                        "AgentOutputRecorder samples queued for recording: %s/%s",
                        qsize,
                        maxsize,
                        extra=extra,
                    )
        next(self._recorders[stream_id])

    def get_recorder(self, stream_id: str) -> Optional[Recorder]:
        """Get the recorder for a stream, if it exists.

        Args:
            stream_id: The stream ID to get recorder for

        Returns:
            The Recorder instance, or None if no recorder exists for this stream
        """
        self.ensure_initialized(stream_id)
        return self._recorders.get(stream_id)

    def has_recorder(self, stream_id: str) -> bool:
        """Check if a recorder exists for a stream.

        Args:
            stream_id: The stream ID to check

        Returns:
            True if a recorder exists for this stream, False otherwise
        """
        return stream_id in self._recorders

    def get_recorders_for_capabilities(self, stream_id: str, capability_names: List[str]) -> List[Recorder]:
        """Get recorders for specified capabilities in a stream.

        This is used by CreateCase capability to collect recorders for coordinated recording.

        Handles two types of recorders:
        1. Agent-level recorders: For capability outputs (all capabilities share one per stream)
        2. DataSource recorders: For input video (stored in capability._dsps)

        Args:
            stream_id: The stream ID to get recorders for
            capability_names: List of capability names to get recorders for

        Returns:
            List containing the Recorder instance(s), or empty list if no recorder exists.
            - For agent-level recording: Returns the shared recorder for this stream
            - For DataSource recording: Returns the capability's _dsps recorder
        """
        recorders = []

        # Check if we have agent-level recorder (for capability outputs)
        self.ensure_initialized(stream_id)
        if stream_id in self._recorders:
            recorders.append(self._recorders[stream_id])

        # Also check for DataSource recorders (for input video)
        # DataSource capabilities have their own _dsps dict
        # get sub graph
        head_node_name = self._stream_graph_paths.get(stream_id)
        if head_node_name is None:
            # Fallback: try to get from thread-local context if available
            try:
                head_node_name = self.pipeline.get_stream()[0].graph_path
            except AttributeError:
                # If thread-local access fails, use the first graph node as default
                graph_items = list(self.pipeline.pipeline_graph._graph.items())
                head_node_name = graph_items[0][0] if graph_items else "default"

        graph_path = self.pipeline.pipeline_graph.get_path(head_node_name)
        for node in graph_path:
            node_name = node.name
            if node_name not in capability_names:
                continue

            capability = node.element
            # Check if this is a DataSource with its own _dsps
            if hasattr(capability, "_dsps") and stream_id in capability._dsps:
                recorders.append(capability._dsps[stream_id])

        if not recorders:
            self.logger.warning(
                f"No recorders found with capabilities {capability_names}. "
                f"Recording may not be enabled for this stream or these capabilities.",
                extra={"stream_id": stream_id},
            )

        return recorders

    def get_all_recorders_for_capabilities(self, capability_names: List[str]) -> List[Recorder]:
        """Get recorders for the supplied capability names across *all* streams."""

        all_recorders: List[Recorder] = []
        seen = set()
        stream_ids = set(self._stream_graph_paths.keys()) | set(self._recorders.keys())

        self.logger.debug(
            "get_all_recorders_for_capabilities: collecting from streams %s for capabilities %s",
            sorted(stream_ids) if stream_ids else "<none>",
            capability_names,
        )

        for stream_id in stream_ids:
            for recorder in self.get_recorders_for_capabilities(stream_id, capability_names):
                marker = id(recorder)
                if marker in seen:
                    continue
                seen.add(marker)
                # Diagnostic: log stream_id of each recorder
                self.logger.debug(
                    "Collected recorder with stream_id=%r, capability_name=%r from stream %s",
                    recorder._stream_id,
                    recorder._capability_name,
                    stream_id,
                )
                all_recorders.append(recorder)

        if not all_recorders:
            self.logger.warning(
                "No recorders found across streams %s for capabilities %s",
                sorted(stream_ids) if stream_ids else "<none>",
                capability_names,
            )

        return all_recorders



import click
import yaml

from .common import _to_pathlib


@click.group("step")
@click.pass_context
def step_group(ctx):
    pass


@step_group.command("create")
@click.option(
    "-n",
    "--name",
    type=str,
    required=True,
)
@click.option(
    "-p",
    "--workflow-id",
    type=str,
    required=True,
)
@click.option(
    "-b",
    "--batch-size",
    type=int,
    required=False,
    default=10,
)
@click.option(
    "--consumable",
    is_flag=True,
)
@click.option(
    "--lockable",
    is_flag=True,
)
@click.option(
    "--capacity",
    type=int,
    required=False,
    default=None,
)
@click.option(
    "--user-ids",
    type=int,
    multiple=True,
)
@click.option(
    "-i",
    "--ids-txt",
    type=click.Path(file_okay=True, writable=False),
    required=True,
    callback=_to_pathlib,
)
@click.pass_context
def create(ctx, name, workflow_id, batch_size, consumable, lockable, capacity, user_ids, ids_txt):
    client = ctx.obj["client"]

    result = client.createImageQueue(return_type=TrainingRunType, id=id)

    result_str = yaml.dump(result.dict())
    print(f"{result_str}")



"""Motion measurement using optical flow."""

import logging
from collections import deque
from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, Union
from uuid import UUID

from highlighter.core.exceptions import require_package

try:
    import matplotlib.pyplot as plt
except ModuleNotFoundError as _:
    plt = None

import numpy as np
from pydantic import BaseModel, model_validator

from highlighter.agent.capabilities import Capability, StreamEvent
from highlighter.agent.capabilities.image_to_scalar._neuflowv2 import (
    NeuFlowV2,
    flow_to_image,
)
from highlighter.client.base_models.annotation import Annotation
from highlighter.client.base_models.entity import Entity
from highlighter.client.base_models.observation import Observation
from highlighter.core import LabeledUUID
from highlighter.core.data_models import DataSample

__all__ = ["MotionMeasurer", "OpticalFlow", "MotionMeasurerCapability"]


DEFAULT_MODEL_PATH = Path.home() / ".cache" / "highlighter" / "models" / "neuflow_sintel.onnx"


class OpticalFlow(BaseModel):
    """
    A standalone optical flow class that calculates motion between consecutive frames.

    All preprocessing (Gaussian blur, resize) is handled inside the ONNX model,
    making this class a thin wrapper around NeuFlowV2.

    Args:
        model_path: Path to the optical flow model file.
        use_gpu: Whether to use GPU for inference.
        blur_kernel_size: Size of Gaussian blur kernel (3, 5, 7) or None to disable blur.
    """

    model_path: Union[str, Path] = DEFAULT_MODEL_PATH
    use_gpu: bool = True
    blur_kernel_size: Optional[int] = 3

    @model_validator(mode="after")
    def init(self):
        self._logger = logging.getLogger(__name__)
        self._estimator = self._load_model()
        self._prev_frame = None
        self._logger.info(f"Initialized OpticalFlow (use_gpu={self.use_gpu}, blur={self.blur_kernel_size})")
        return self

    def _load_model(self) -> NeuFlowV2:
        """Load the NeuFlowV2 optical flow model."""
        return NeuFlowV2(str(self.model_path), use_gpu=self.use_gpu, blur_kernel_size=self.blur_kernel_size)

    def update(self, image: np.ndarray) -> np.ndarray:
        """
        Update with a new image and return optical flow vectors.

        Args:
            image: Input image (H, W, 3) uint8 array.

        Returns:
            Flow vectors (H, W, 2) float32 array.
        """
        # Ensure uint8
        if image.dtype != np.uint8:
            image = (image * 255).astype(np.uint8) if image.max() <= 1.0 else image.astype(np.uint8)

        if self._prev_frame is None:
            self._prev_frame = image
            h, w = image.shape[:2]
            return np.zeros((h, w, 2), dtype=np.float32)

        flow_vectors = self._estimator(self._prev_frame, image)
        self._prev_frame = image

        return flow_vectors

    def reset(self):
        """Reset the optical flow state."""
        self._prev_frame = None

    @staticmethod
    def draw_flow(curr_frame: np.ndarray, flow: np.ndarray, step: int = 10) -> np.ndarray:
        """
        Draw optical flow vectors on a frame.

        Uses the Middlebury color scheme via flow_to_image and overlays on the frame.

        Args:
            curr_frame: Current frame (H, W, 3) uint8 array.
            flow: Optical flow (H, W, 2) float32 array.
            step: Grid step for sampling flow vectors.

        Returns:
            Frame with flow visualization (H, W, 3) uint8 array.
        """
        # Create flow color image
        flow_img = flow_to_image(flow)

        # Blend with original frame
        alpha = 0.6
        blended = (alpha * flow_img + (1 - alpha) * curr_frame).astype(np.uint8)

        return blended


class MotionMeasurer(BaseModel):
    """
    Compute motion measurement from video frames via optical flow.

    Usage:
        mm = MotionMeasurer(score_type=["mean", "median"], window_size=10)
        for idx, frame in enumerate(frame_generator):
            scores = mm.update(frame, idx)
            # {'mean': 0.23, 'median': 0.19}
    """

    score_type: List[Literal["sum", "mean", "median"]] = ["mean"]
    window_size: int = 20
    history_size: int = 40

    # OpticalFlow Args
    model_path: Union[str, Path] = DEFAULT_MODEL_PATH
    use_gpu: bool = True
    blur_kernel_size: Optional[int] = 3

    @model_validator(mode="after")
    def init(self):
        self._optical_flow_model = OpticalFlow(
            model_path=self.model_path,
            use_gpu=self.use_gpu,
            blur_kernel_size=self.blur_kernel_size,
        )

        self._score_names = ["sum", "mean", "median"]
        self._score_idxs = [self._score_names.index(a) for a in self.score_type]
        self._motion_averages = {
            self._score_names[i]: MovingAverage(self.window_size) for i in self._score_idxs
        }
        self._frame_idxs = deque(maxlen=self.history_size)
        self._motion_averages_history = {
            self._score_names[i]: deque(maxlen=self.history_size) for i in self._score_idxs
        }
        self._flow_vectors = None
        return self

    def _update_motion_scores(self, flow_vectors: np.ndarray):
        u, v = flow_vectors[..., 0], flow_vectors[..., 1]
        magnitude = np.sqrt(u**2 + v**2)
        motion_scores = (
            float(np.sum(magnitude)),
            float(np.mean(magnitude)),
            float(np.median(magnitude)),
        )

        for i in self._score_idxs:
            self._motion_averages[self._score_names[i]].update(motion_scores[i])

        for k, v in self._motion_averages.items():
            self._motion_averages_history[k].append(v.get_average())

    def get_latest_flow_vectors(self):
        return self._flow_vectors

    def update(self, image: np.ndarray, frame_idx: int) -> Dict[str, float]:
        """
        Process a new video frame and update motion statistics.

        Parameters
        ----------
        image : np.ndarray
            Current video frame (H, W, 3) uint8 array.
        frame_idx : int
            Frame index for tracking.

        Returns
        -------
        Dict[str, float]
            Mapping from score name to moving-average value.
        """
        self._frame_idxs.append(frame_idx)
        flow_vectors = self._optical_flow_model.update(image)
        self._flow_vectors = flow_vectors

        self._update_motion_scores(flow_vectors)

        return {k: v[-1] for k, v in self._motion_averages_history.items()}

    @require_package(plt, "matplotlib", "matplotlib")
    def get_motion_score_fig(self, select: Optional[List[str]] = None, width: int = 600, height: int = 550):
        plt.style.use("dark_background")

        if select is None:
            select = list(self._motion_averages.keys())

        assert all([s in self._motion_averages for s in select])

        fig, axies = plt.subplots(len(select), 1, figsize=(width / 100, height / 100), dpi=100)
        if not isinstance(axies, np.ndarray):
            axies = [axies]

        plot_colors = {
            "sum": "#00BFFF",
            "mean": "#00FF7F",
            "median": "#FFD700",
        }
        plot_colors = {k: v for k, v in plot_colors.items() if k in select}

        for i, (key, color) in enumerate(plot_colors.items()):
            ax = axies[i]
            scores = self._motion_averages_history[key]

            ax.plot(self._frame_idxs, scores, color=color, linewidth=2, alpha=0.9)
            ax.fill_between(self._frame_idxs, scores, alpha=0.3, color=color)

            ax.set_title(
                f"Movement Score ({key.title()})",
                fontsize=20,
                color="white",
                fontweight="bold",
            )
            ax.set_ylabel(key.title(), fontsize=20, color="white")
            ax.grid(True, alpha=0.4, color="gray", linestyle="-", linewidth=0.8)
            ax.set_facecolor("black")
            ax.tick_params(colors="white", labelsize=20)

        axies[-1].set_xlabel("Frame", fontsize=7, color="white")

        fig.patch.set_facecolor("black")
        fig.patch.set_alpha(0.9)
        plt.tight_layout()
        return fig


class MovingAverage:
    def __init__(self, window_size):
        if window_size <= 0:
            raise ValueError("Window size must be positive")
        self.window_size = window_size
        self.values = []
        self.sum = 0.0

    def update(self, value):
        self.values.append(value)
        self.sum += value

        if len(self.values) > self.window_size:
            self.sum -= self.values.pop(0)

        return self.get_average()

    def get_average(self):
        if not self.values:
            return 0.0
        return self.sum / len(self.values)


class MotionMeasurerCapability(Capability):
    class InitParameters(Capability.InitParameters):
        attribute_id: UUID
        object_class_id: UUID | str
        model_path: str = DEFAULT_MODEL_PATH
        score_type: Literal["mean", "sum", "median"] = "mean"
        window_size: int = 20
        use_gpu: bool = False
        blur_kernel_size: Optional[int] = 3

    class StreamParameters(InitParameters):
        entity_id: UUID
        crop: Optional[Tuple[float, float, float, float]] = None

    def __init__(self, context):
        super().__init__(context)

        if "|" in self.init_parameters.object_class_id:
            self.object_class_id = LabeledUUID.from_str(self.init_parameters.object_class_id)
        else:
            self.object_class_id = LabeledUUID.from_str(f"{self.init_parameters.object_class_id}|-")

    def start_stream(self, stream, stream_id, use_create_frame=True):
        kwargs = {"score_type": [self.init_parameters.score_type]}
        kwargs["model_path"] = self.init_parameters.model_path
        kwargs["use_gpu"] = self.init_parameters.use_gpu
        kwargs["blur_kernel_size"] = self.init_parameters.blur_kernel_size
        kwargs["window_size"] = self.init_parameters.window_size
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        stream.variables["_motion_score_model"] = MotionMeasurer(**kwargs)
        return super().start_stream(stream, stream_id, use_create_frame=use_create_frame)

    def stop_stream(self, stream, stream_id):
        return super().stop_stream(stream, stream_id)

    def process_frame(
        self, stream, data_samples: List[DataSample], **kwargs
    ) -> Tuple[StreamEvent, Union[Dict, str]]:
        parameters = self.stream_parameters(stream.stream_id)
        if len(data_samples) > 1:
            self.logger.warning(
                f"Expected only a single DataFile, got {len(data_samples)}. Processing data_samples[0] only."
            )

        ds = data_samples[0]
        w, h = ds.wh

        if parameters.crop is not None:
            if isinstance(parameters.crop[0], float):
                x0 = int(parameters.crop[0] * w)
                y0 = int(parameters.crop[1] * h)
                x1 = int(parameters.crop[2] * w)
                y1 = int(parameters.crop[3] * h)
                anno_tuple = (x0, y0, x1, y1)
            else:
                anno_tuple = parameters.crop
            img = ds.crop_content([parameters.crop])[0]
        else:
            anno_tuple = (0, 0, w, h)
            img = ds.content

        score = stream.variables["_motion_score_model"].update(img, ds.media_frame_index)[
            parameters.score_type
        ]
        attribute_id = LabeledUUID.from_str(f"{parameters.attribute_id}|motion_score")
        score_obs = Observation.make_scalar_observation(
            score,
            attribute_id,
            occurred_at=ds.recorded_at,
            pipeline_element_name=self.name,
            frame_id=ds.media_frame_index,
        )
        obj_obs = Observation.make_object_class_observation(
            object_class_uuid=self.object_class_id,
            object_class_value=self.object_class_id.label,
            confidence=1.0,
            occurred_at=ds.recorded_at,
            pipeline_element_name=self.name,
            frame_id=ds.media_frame_index,
        )
        a = Annotation.from_left_top_right_bottom_box(
            anno_tuple,
            1.0,
            data_sample=ds,
            observations=[score_obs, obj_obs],
        )

        entity = Entity(id=parameters.entity_id, annotations=[a])
        entities = {parameters.entity_id: entity}

        for entity_id, entity in entities.items():
            for anno in entity.annotations:
                for obs in anno.observations:
                    if hasattr(obs, "attribute_id") and "motion_score" in str(obs.attribute_id):
                        self.logger.verbose(
                            f"motion_measurer output: entity={entity_id}, frame={obs.datum_source.frame_id}, attr={obs.attribute_id}, value={obs.value}, type={type(obs.value)}"
                        )

        return StreamEvent.OKAY, {"entities": entities}



from __future__ import annotations

"""Shared, thread-safe resource cache with in-flight coordination.

Example
-------
>>> cache = SharedResourceCache[str](maxsize=2)
>>> def build():
...     return "value"
>>> cache.get_or_create("key", build)
'value'
>>> cache.get_or_create("key", build)
'value'
>>>
>>> # Example: sharing a heavy ONNX Runtime session
>>> # from highlighter.predictors.ort.session import build_session  # pseudo-helper
>>> # session = cache.get_or_create(session_key, build_session)
"""

import threading
from collections import OrderedDict
from typing import Callable, Dict, Generic, Hashable, Optional, TypeVar

T = TypeVar("T")


class _Inflight(Generic[T]):
    __slots__ = ("event", "error", "value")

    def __init__(self) -> None:
        self.event = threading.Event()
        self.error: Optional[BaseException] = None
        self.value: Optional[T] = None


class SharedResourceCache(Generic[T]):
    """Thread-safe cache for shared heavyweight resources.

    - Ensures only one builder runs per key at a time.
    - Waiters block until the in-flight build completes.
    - LRU eviction enforces a maximum cache size (maxsize <= 0 disables eviction).
    """

    def __init__(self, maxsize: int = 128) -> None:
        self._cache: "OrderedDict[Hashable, T]" = OrderedDict()
        self._inflight: Dict[Hashable, _Inflight[T]] = {}
        self._lock = threading.RLock()
        self._maxsize = maxsize

    def _evict_if_needed(self, maxsize: int) -> None:
        if maxsize <= 0:
            return
        while len(self._cache) > maxsize:
            self._cache.popitem(last=False)

    def get_or_create(
        self,
        key: Hashable,
        create_fn: Callable[[], T],
        *,
        maxsize: Optional[int] = None,
    ) -> T:
        size = self._maxsize if maxsize is None else maxsize
        with self._lock:
            cached = self._cache.get(key)
            if cached is not None:
                self._cache.move_to_end(key)
                return cached

            inflight = self._inflight.get(key)
            if inflight is None:
                inflight = _Inflight[T]()
                self._inflight[key] = inflight
                creator = True
            else:
                creator = False

        if not creator:
            inflight.event.wait()
            if inflight.error:
                raise RuntimeError("Shared resource creation failed.") from inflight.error
            if inflight.value is None:
                raise RuntimeError("Shared cache creation finished without a value.")
            return inflight.value

        try:
            value = create_fn()
        except Exception as exc:
            with self._lock:
                inflight.error = exc
                self._inflight.pop(key, None)
                inflight.event.set()
            raise

        with self._lock:
            self._cache[key] = value
            self._evict_if_needed(int(size))

            inflight.value = value
            self._inflight.pop(key, None)
            inflight.event.set()

        return value

    def clear(self) -> None:
        with self._lock:
            self._cache.clear()

    def keys(self) -> list[Hashable]:
        with self._lock:
            return list(self._cache.keys())



import shutil
from pathlib import Path

from sqlalchemy.engine import Engine
from sqlmodel import Session, SQLModel, create_engine
from sqlmodel.pool import StaticPool

__all__ = [
    "TestDatabase",
    "RuntimeDatabaseTemplate",
]


class TestDatabase:
    """The Highlighter agent database

    Each instance creates its own in-memory SQLite database to ensure test isolation.
    """

    def __init__(self):
        # Create a fresh in-memory database for each instance
        # This ensures tests don't interfere with each other
        self.engine: Engine = create_engine(
            "sqlite://", connect_args={"check_same_thread": False}, poolclass=StaticPool
        )
        SQLModel.metadata.create_all(self.engine)

    def get_session(self):
        return Session(self.engine)


class RuntimeDatabaseTemplate:
    """Create a single migrated SQLite DB that tests can restore quickly.

    Pytest spins up many ``Runtime`` instances and each one initialises
    ``highlighter.core.database.Database`` which, in turn, runs Alembic
    migrations. Applying all migrations for every test case is slow and noisy.

    ``RuntimeDatabaseTemplate`` initialises the on-disk database once per test
    session, captures the fully migrated file as ``database.template.sqlite``
    and provides a cheap ``reset()`` helper that clones the template back into
    ``database.sqlite`` before each test. That keeps the schema current while
    avoiding repeated calls to Alembic.
    """

    def __init__(self, cache_dir: Path):
        self._cache_dir = Path(cache_dir)
        self._db_dir = self._cache_dir / "agents" / "db"
        self._db_path = self._db_dir / "database.sqlite"
        self._template_path = self._db_dir / "database.template.sqlite"

    def prepare(self) -> None:
        """Ensure the template file exists and mirrors the latest migrations."""

        from highlighter.core.database.database import (
            Database,  # Local import to avoid cycles
        )

        self._db_dir.mkdir(parents=True, exist_ok=True)

        db = None
        try:
            db = Database()
            # ``Database`` runs Alembic migrations + metadata.create_all(...) on init
        finally:
            if db is not None:
                db.close()

        if not self._db_path.exists():
            raise RuntimeError(f"Expected database at {self._db_path} after migrations ran")

        shutil.copy2(self._db_path, self._template_path)

    def reset(self) -> None:
        """Restore the working DB from the pre-migrated template."""

        if not self._template_path.exists():
            raise RuntimeError("Runtime DB template missing – call prepare() first")

        self._db_dir.mkdir(parents=True, exist_ok=True)

        # Replace the working database with a pristine copy for the next test
        if self._db_path.exists():
            self._db_path.unlink()
        shutil.copy2(self._template_path, self._db_path)
